# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo
import math

from sdaps import model
from sdaps import defs


LINE_WIDTH = 25.4/72
MIN_FREETEXT_SIZE = 4.0

_LEFT = 1
_RIGHT = 2
_TOP = 3
_BOTTOM = 4


def inner_box(cr, x, y, width, height):
    line_width = cr.get_line_width()

    cr.rectangle(x + line_width / 2.0, y + line_width / 2.0,
                 width - line_width, height - line_width)

def ellipse(cr, x, y, width, height):
    cr.save()

    cr.translate(x + width / 2.0, y + height / 2.0)

    line_width = cr.get_line_width()

    cr.scale(width / 2.0, height / 2.0)
    cr.arc(0, 0, 1.0, 0, 2*math.pi)
    cr.close_path()

    # Restore old matrix (without removing the current path)
    cr.restore()

def inner_ellipse(cr, x, y, width, height):
    cr.save()

    cr.translate(x + width / 2.0, y + height / 2.0)

    line_width = cr.get_line_width()

    cr.scale((width - line_width) / 2.0, (height - line_width) / 2.0)
    cr.arc(0, 0, 1.0, 0, 2*math.pi)
    cr.close_path()

    # Restore old matrix (without removing the current path)
    cr.restore()

class Questionnaire(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'gui'
    obj_class = model.questionnaire.Questionnaire

    def draw(self, cr, page_number):
        for qobject in self.obj.qobjects:
            qobject.gui.draw(cr, page_number)

    def find_box(self, page_number, x, y):
        for qobject in self.obj.qobjects:
            result = qobject.gui.find_box(page_number, x, y)
            if result:
                return result
        return None

    def find_edge(self, page_number, x, y, tollerance_x, tollerance_y):
        for qobject in self.obj.qobjects:
            result = qobject.gui.find_edge(page_number, x, y, tollerance_x, tollerance_y)
            if result:
                return result
        return None


class QObject(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'gui'
    obj_class = model.questionnaire.QObject

    def draw(self, cr, page_number):
        pass

    def find_box(self, page_number, x, y):
        pass

    def find_edge(self, page_number, x, y, tollerance_x, tollerance_y):
        return None


class Question(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'gui'
    obj_class = model.questionnaire.Question

    def draw(self, cr, page_number):
        # Does the sheet contain this question?
        if page_number == self.obj.page_number:
            # iterate over boxes
            for box in self.obj.boxes:
                box.gui.draw(cr)

    def find_box(self, page_number, x, y):
        # Does the sheet contain this question?
        if page_number == self.obj.page_number:
            # iterate over boxes
            for box in self.obj.boxes:
                result = box.gui.find(x, y)
                if result is not None:
                    return result
        return None

    def find_edge(self, page_number, x, y, tollerance_x, tollerance_y):
        # Does the sheet contain this question?
        if page_number == self.obj.page_number:
            # iterate over boxes
            for box in self.obj.boxes:
                result = box.gui.find_edge(x, y, tollerance_x, tollerance_y)
                if result is not None:
                    return result
        return None


class Box(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'gui'
    obj_class = model.questionnaire.Checkbox

    def draw(self, cr):
        cr.set_fill_rule(cairo.FILL_RULE_EVEN_ODD)

        cr.set_source_rgba(0.57, 1.0, 0.0, 0.5)
        cr.set_line_width(self.obj.lw)

        inner_box(cr, self.obj.data.x, self.obj.data.y, self.obj.data.width, self.obj.data.height)
        cr.stroke()

    def find(self, x, y):
        if self.obj.data.x < x < self.obj.data.x + self.obj.data.width:
            if self.obj.data.y < y < self.obj.data.y + self.obj.data.height:
                return self.obj
        return None

    def find_edge(self, x, y, tollerance_x, tollerance_y):
        return None


class Checkbox(Box, metaclass=model.buddy.Register):

    name = 'gui'
    obj_class = model.questionnaire.Checkbox

    def find(self, x, y):
        if self.obj.data.x - 2*LINE_WIDTH < x < self.obj.data.x + self.obj.data.width + 2*LINE_WIDTH:
            if self.obj.data.y - 2*LINE_WIDTH < y < self.obj.data.y + self.obj.data.height + 2*LINE_WIDTH:
                return self.obj
        return None


    def draw(self, cr):
        cr.save()

        cr.set_fill_rule(cairo.FILL_RULE_EVEN_ODD)

        if self.obj.data.quality < 0.5:
            cr.save()
            cr.set_line_width(2*LINE_WIDTH)
            cr.set_source_rgba(1.0, 0.0, 0.2, 0.6)

            if self.obj.form == "box":
                inner_box(cr, self.obj.data.x - 4*LINE_WIDTH, self.obj.data.y - 4*LINE_WIDTH,  self.obj.data.width+8*LINE_WIDTH, self.obj.data.height+8*LINE_WIDTH)
                cr.stroke()
            elif self.obj.form == "ellipse":
                inner_ellipse(cr, self.obj.data.x - 4*LINE_WIDTH, self.obj.data.y - 4*LINE_WIDTH,  self.obj.data.width+8*LINE_WIDTH, self.obj.data.height+8*LINE_WIDTH)
                cr.stroke()

            cr.restore()

        cr.set_source_rgba(0.57, 1.0, 0.0, 0.5)
        cr.set_line_width(self.obj.lw)

        if self.obj.data.state:
            if self.obj.form == "box":
                cr.rectangle(self.obj.data.x - 2*LINE_WIDTH, self.obj.data.y - 2*LINE_WIDTH, self.obj.data.width + 4*LINE_WIDTH, self.obj.data.height + 4*LINE_WIDTH)
                cr.fill()
            elif self.obj.form == "ellipse":
                ellipse(cr, self.obj.data.x - 2*LINE_WIDTH, self.obj.data.y - 2*LINE_WIDTH, self.obj.data.width + 4*LINE_WIDTH, self.obj.data.height + 4*LINE_WIDTH)
                cr.fill()
        else:
            if self.obj.form == "box":
                inner_box(cr, self.obj.data.x, self.obj.data.y, self.obj.data.width, self.obj.data.height)
                cr.stroke()
            elif self.obj.form == "ellipse":
                inner_ellipse(cr, self.obj.data.x, self.obj.data.y, self.obj.data.width, self.obj.data.height)
                cr.stroke()

        cr.restore()


class Textbox(Box, metaclass=model.buddy.Register):

    name = 'gui'
    obj_class = model.questionnaire.Textbox

    def draw(self, cr):
        cr.save()

        cr.set_fill_rule(cairo.FILL_RULE_EVEN_ODD)

        cr.set_source_rgba(0.57, 1.0, 0.0, 0.5)

        if self.obj.data.state:
            cr.set_line_width(2 * LINE_WIDTH + self.obj.lw)
        else:
            cr.set_line_width(self.obj.lw)

        inner_box(cr, self.obj.data.x, self.obj.data.y, self.obj.data.width, self.obj.data.height)
        cr.stroke()

        cr.restore()

    def find_edge(self, x, y, tollerance_x, tollerance_y):
        if self.obj.data.x - tollerance_x <= x and \
           self.obj.data.x + tollerance_x >= x and \
           self.obj.data.y <= y and \
           self.obj.data.y + self.obj.data.height >= y:
            return(self.obj, _LEFT)

        if self.obj.data.x + self.obj.data.width - tollerance_x <= x and \
           self.obj.data.x + self.obj.data.width + tollerance_x >= x and \
           self.obj.data.y <= y and \
           self.obj.data.y + self.obj.data.height >= y:
            return(self.obj, _RIGHT)

        if self.obj.data.y - tollerance_y <= y and \
           self.obj.data.y + tollerance_y >= y and \
           self.obj.data.x <= x and \
           self.obj.data.x + self.obj.data.width >= x:
            return(self.obj, _TOP)

        if self.obj.data.y + self.obj.data.height - tollerance_y <= y and \
           self.obj.data.y + self.obj.data.height + tollerance_y >= y and \
           self.obj.data.x <= x and \
           self.obj.data.x + self.obj.data.width >= x:
            return(self.obj, _BOTTOM)

    def move_edge(self, side, x, y):
        if side == _LEFT:
            x = max(x, defs.corner_mark_left)
            new_width = max(MIN_FREETEXT_SIZE, self.obj.data.width + self.obj.data.x - x)
            x = self.obj.data.x + (self.obj.data.width - new_width)

            self.obj.data.width = new_width
            self.obj.data.x = x
        elif side == _RIGHT:
            x = min(x, self.obj.question.questionnaire.survey.defs.paper_width - defs.corner_mark_right)
            new_width = max(MIN_FREETEXT_SIZE, x - self.obj.data.x)
            new_width = min(new_width, self.obj.data.x + self.obj.data.width)

            self.obj.data.width = new_width
        elif side == _TOP:
            y = max(y, defs.corner_mark_top)
            new_height = max(MIN_FREETEXT_SIZE, self.obj.data.height + self.obj.data.y - y)
            new_y = self.obj.data.y + (self.obj.data.height - new_height)

            self.obj.data.height = new_height
            self.obj.data.y = new_y
        elif side == _BOTTOM:
            y = min(y, self.obj.question.questionnaire.survey.defs.paper_height - defs.corner_mark_bottom)
            new_height = max(MIN_FREETEXT_SIZE, y - self.obj.data.y)

            self.obj.data.height = new_height





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This modules contains low level image processing functions. These functions
are implemented in C for speed reasons. Usually one will not need to use these
directly, instead modules like "recognize" or "surface" use them to load and
analyze the image data.
"""

import os
import sys

from sdaps import paths
from sdaps import defs
from sdaps import log

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

if paths.local_run:
    # image.so liegt in lib_build_dir/image/
    __path__.append(os.path.join(paths.lib_build_dir, 'image'))

# If SDAPS is installed, then the image.so file is in the current directory.
# Simply importing it without changes to the paths will work.

try:
    from .image import *
except ImportError as e:
    print(e)
    log.error(_("It appears you have not build the C extension. Please run \"./setup.py build\" in the toplevel directory."))
    sys.exit(1)

set_magic_values(defs.corner_mark_min_length,
                 defs.corner_mark_max_length,
                 defs.image_line_width,
                 defs.corner_mark_search_distance,
                 defs.image_line_coverage)




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, 2013, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from sdaps import log
import re

from sdaps import defs
import subprocess
import os

try:
    from sdaps.utils.latexmap import mapping
except ImportError:
    mapping = {}
    log.warn(_('The latex character map is missing! Please build it using the supplied tool (create-latexmap.py).'))

# Add some more mappings
# NBSP
mapping['~'] = ' '


re_latex_to_unicode_mapping = {}
for token, replacement in mapping.items():
    regexp = re.compile('%s(?=^w|})' % re.escape(token))
    re_latex_to_unicode_mapping[regexp] = replacement

# Regular expressions don't work really, but we replace a single string anyways
unicode_to_latex_mapping = {}
for token, replacement in mapping.items():
    unicode_to_latex_mapping[replacement] = "{%s}" % token


def latex_to_unicode(string):
    string = str(string)
    for regexp, replacement in re_latex_to_unicode_mapping.items():
        string, count = regexp.subn(replacement, string)

    def ret_char(match):
        return match.group('char')
    string, count = re.subn(r'\\IeC {(?P<char>.*?)}', ret_char, string)
    return string

def unicode_to_latex(string):
    string = str(string)
    for char, replacement in unicode_to_latex_mapping.items():
        string = string.replace(char, replacement)

    # The returned string may still contain unicode characters if
    # the user is using xelatex. But in that case, the remapping is not
    # needed anyway.
    # However, it could also mean that the mapping needs to be updated.
    try:
        string.encode('ascii')
    except UnicodeEncodeError:
        log.warn(_("Generated string for LaTeX contains unicode characters. This may not work correctly and could mean the LaTeX character map needs to be updated."))
    return string

def quote_braces(string):
    return string.replace('{', '\\{').replace('}', '\\}')

def write_override(survey, optfile, draft=False, questionnaire_ids=None):
    latex_override = open(optfile, 'w')

    if questionnaire_ids:
        quoted_ids = [quote_braces(str(id)) for id in questionnaire_ids]
    else:
        quoted_ids = []

    latex_override.write('''
%% This file exists to force the latex document into "final" mode.
%% It is parsed after the setup phase of the SDAPS class.

%% Old class vs. new class
\\ifcsname @PAGEMARKtrue\endcsname
    \\setcounter{surveyidlshw}{%(survey_id_lshw)i}
    \\setcounter{surveyidmshw}{%(survey_id_mshw)i}
    \\def\\surveyid{%(survey_id)i}
    %(noglobalid)s\\def\\globalid{%(global_id)s}
    \\@STAMPtrue
    \\@PAGEMARKtrue
    \\@sdaps@draft%(draft)s
    \\def\\questionnaireids{%(qids_old)s}
\\else
  \\group_begin:
    \\def\\setoptions#1#2#3{
      \\tl_gset:Nn \\g_sdaps_survey_id_tl { #1 }
      %(noglobalid)s\\tl_gset:Nn \\g_sdaps_global_id_tl { #2 }
      \\seq_gset_from_clist:Nn \\g_sdaps_questionnaire_ids_seq { #3 }
    }
    \\bool_gset_%(draft)s:N \g_sdaps_draft_bool

    \\ExplSyntaxOff
      \\setoptions{%(survey_id)i}{%(global_id)s}{%(qids)s}
    \\ExplSyntaxOn
  \group_end:
\\fi
''' % {
            'survey_id' : survey.survey_id,
            'survey_id_lshw' : (survey.survey_id % (2 ** 16)),
            'survey_id_mshw' : (survey.survey_id / (2 ** 16)),
            'draft' : 'true' if draft else 'false',
            'noglobalid' : '%' if draft else '',
            'global_id' : quote_braces(survey.global_id) if survey.global_id is not None else '',
            'qids_old' : '{' + '},{'.join(quoted_ids) + '}' if quoted_ids else '{NONE}',
            'qids' : '{' + '},{'.join(quoted_ids) + '}' if quoted_ids else '{}',
        })
    latex_override.close()

# This is a list, because the order is relevant!
ascii_to_latex = [
    ('{', '\\{'),
    ('}', '\\}'),
    ('\\', '{\\textbackslash}'),
    ('%', '\\%'),
    ('$', '\\$'),
    ('_', '\\_'),
    ('|', '{\\textbar}'),
    ('>', '{\\textgreater}'),
    ('<', '{\\textless}'),
    ('&', '\\&'),
    ('#', '\#'),
    ('^', '\\^{}'),
    ('~', '\\~{}'),
    ('"', '\\"{}'),
]

def raw_unicode_to_latex(string):
    """In addition to converting all unicode characters to LaTeX expressions
    this function also replaces some characters like newlines with their LaTeX
    equvivalent."""

    # We need to quote any special character (or replace it)
    for char, replacement in ascii_to_latex:
        string = string.replace(char, replacement)

    string = unicode_to_latex(string)

    string = str(string)

    # Replace many newlines with a paragraph marker
    string, count = re.subn('\n\n+', '\u2029', string, flags=re.MULTILINE)

    # Replace single newline with \\+newline
    string = string.replace('\n', '\\\\\n')

    # And remove the paragraph marker again (insert two newlines)
    string = string.replace('\u2029', '\n\n')

    return string.encode('ascii')

def run_engine(engine, texfile, cwd, inputs=[]):
    def _preexec_fn():
        if defs.latex_preexec_hook is not None:
            defs.latex_preexec_hook()

        inputs.extend(p for p in os.environ.get('TEXINPUTS', '').split(':') if p not in ['.', ''])

        if inputs:
            os.environ['TEXINPUTS'] = ':'.join(['.'] + inputs + [''])

    verbose = os.environ.get('VERBOSE', '0')
    if verbose in ['1', 'y']:
        mode = 'nonstopmode'
    else:
        mode = 'batchmode'
    subprocess.call([engine, '-halt-on-error',
                     '-interaction', mode, texfile],
                    cwd=cwd,
                    preexec_fn=_preexec_fn)


def compile(engine, texfile, cwd, inputs=[]):
    run_engine(engine, texfile, cwd, inputs)
    run_engine(engine, texfile, cwd, inputs)
    run_engine(engine, texfile, cwd, inputs)
    run_engine(engine, texfile, cwd, inputs)





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''Some path values, used to find and load program components.
'''

import os


import gettext
import locale


local_run = False

# required if local_run == True
build_dir = str()

# required if local_run == True
lib_build_dir = str()

# required if local_run == True
source_dir = str()

# required if local_run == False
prefix = str()


def init(local_run_value, package_path):
    '''Initialize path values for sdaps
    '''
    global local_run, build_dir, lib_build_dir, source_dir, prefix

    # Initialize local_run
    local_run = local_run_value

    base_dir = os.path.split(os.path.abspath(package_path))[0]

    if local_run:
        source_dir = base_dir

        from pkg_resources import get_build_platform
        from distutils.sysconfig import get_python_version

        # Initialize gettext
        init_gettext(os.path.join(
            base_dir,
            'build',
            'mo'))

        # Initialize build_dir
        build_dir = os.path.join(base_dir, 'build', 'share', 'sdaps')

        # Initialize build_dir
        lib_build_dir = os.path.join(
            base_dir,
            'build', 'lib.%s-%s' % (get_build_platform(), get_python_version()),
            'sdaps')
    else:
        # Look for the data in the parent directories
        path = base_dir
        while True:
            if os.path.exists(os.path.join(path, 'share', 'sdaps')):
                prefix = path
                break
            new_path = os.path.split(path)[0]
            assert not path == new_path, "could not find locales" # Wir wären oben angekommen
            path = new_path

        # Initialize gettext
        init_gettext(os.path.join(prefix, 'share', 'locale'))


def init_gettext(locale_dir):
    '''Initialize gettext using the given directory containing the l10n data.
    '''
    gettext.bindtextdomain('sdaps', locale_dir)

    if hasattr(gettext, 'bind_textdomain_codeset'):
        gettext.bind_textdomain_codeset('sdaps', 'UTF-8')
        gettext.textdomain('sdaps')
    if hasattr(locale, 'bind_textdomain_codeset'):
        locale.bindtextdomain('sdaps', locale_dir)
        locale.bind_textdomain_codeset('sdaps', 'UTF-8')
        locale.textdomain('sdaps')




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2012, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Poppler
import cairo
from . import buddies
import os.path

def annotate(survey, infile=None, outfile=None):
    if infile is None:
        infile = 'file://' + os.path.abspath(survey.path('questionnaire.pdf'))
    else:
        infile = 'file://' + os.path.abspath(infile)

    if outfile is None:
        outfile = survey.path('annotated_questionnaire.pdf')

    pdf = Poppler.Document.new_from_file(infile, None)

    width, height = pdf.get_page(0).get_size()

    output = cairo.PDFSurface(outfile, 2*width, 2*height)

    cr = cairo.Context(output)

    for p in range(survey.questionnaire.page_count):
        pdf.get_page(p).render_for_printing(cr)

        cr.save()
        # Use mm space in here.
        cr.scale(72.0 / 25.4, 72.0 / 25.4)

        layout_info = dict()
        layout_info['twidth'] = width / 72.0 * 25.4 - 20 # mm
        layout_info['xshift'] = width / 72.0 * 25.4 + 10 # mm
        layout_info['ypos'] = 15 # mm
        layout_info['font'] = "Sans 3"
        layout_info['boxfont'] = "Sans 2"

        # And, render the questions on top (1 based page numbers here)
        survey.questionnaire.annotate.draw(cr, p + 1, layout_info)

        cr.restore()

        cr.show_page()





#!/usr/bin/env python3
# -*- coding: utf8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright (C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright (C) 2015, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os
import cairo
import math

# Use the following and local_run=True below to run without installing SDAPS
#sys.path.append(os.path.join(os.path.dirname(sys.argv[0]), '..'))

import sdaps
#sdaps.init(local_run=True)
sdaps.init()

from sdaps import model
from sdaps import image
from sdaps import matrix

survey = model.survey.Survey.load(sys.argv[1])
counter = 1


def ellipse(cr, x, y, width, height):
    cr.save()

    cr.translate(x + width / 2.0, y + height / 2.0)

    line_width = cr.get_line_width()

    cr.scale(width / 2.0, height / 2.0)
    cr.arc(0, 0, 1.0, 0, 2*math.pi)
    cr.close_path()

    # Restore old matrix (without removing the current path)
    cr.restore()



def generate_pdf():

    global counter
    global survey

    questionnaire = survey.questionnaire
    sheet = survey.sheet

    # Use the questionnaire ID, if not there, give up?
    if sheet.questionnaire_id is not None:
        pdf_name = 'overlay-%s.pdf' % sheet.questionnaire_id
    else:
        pdf_name = 'overlay-%04i.pdf' % counter
        counter += 1

    surface = cairo.PDFSurface(pdf_name, survey.defs.paper_width * 72 / 25.4, survey.defs.paper_height * 72 / 25.4)
    cr = cairo.Context(surface)

    # Transform to mm space
    cr.scale(72 / 25.4, 72 / 25.4)

    for p in range(questionnaire.page_count):
        # 1 based page numbers
        p += 1

        # Get the current image
        img = sheet.get_page_image(p)
        if img is None:
            print("no image")
            cr.show_page()
            continue

        # Render the image on the background, we take the easy way and paint it
        # correctly. i.e. so that the markings on top are at the expected
        # positions. It would likely be faster to only scale the image instead.
        # (And not much harder, just needs the correct transformation for the
        # other markings)
        cr.save()
        cr.transform(img.matrix.px_to_mm())
        cr.set_source_rgb(0, 0, 0)
        cr.mask_surface(img.surface.load_uncached(), 0, 0)
        cr.restore()

        # Now render all checked checkboxes, for now inline here.
        for qobj in questionnaire.qobjects:
            for box in qobj.boxes:
                if box.data.state:
                    if box.page_number != p:
                        continue

                    if not isinstance(box, model.questionnaire.Checkbox):
                        continue

                    # 1pt line width
                    cr.set_line_width(1 / 25.4)

                    x, y, width, height = box.data.x, box.data.y, box.data.width, box.data.height
                    x -= width / math.sqrt(2) / 2
                    y -= height / math.sqrt(2) / 2
                    width += width / math.sqrt(2)
                    height += height / math.sqrt(2)

                    ellipse(cr, x, y, width, height)
                    cr.set_source_rgb(1, 0, 0)
                    cr.stroke()

        cr.show_page()

    del cr
    surface.flush()
    del surface

survey.iterate_progressbar(generate_pdf)




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008,2018, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import struct

import json
import sqlite3
import weakref
from contextlib import closing
 
from . import db
from . import questionnaire
from .sheet import Sheet
from sdaps import defs
from sdaps import log

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

valid_styles = ['classic', 'code128', 'custom', 'qr']
valid_checkmodes = ['checkcorrect', 'check', 'fill']

_db_schema = """
CREATE TABLE surveys (
    json TEXT
);
CREATE TRIGGER survey_delete AFTER DELETE ON surveys FOR EACH ROW
  BEGIN
    DELETE FROM sheets WHERE survey_rowid = OLD.rowid;
  END;


CREATE TABLE sheets (
    survey_rowid REFERENCES surveys(rowid) NOT NULL,
    sort INTEGER(8),

    json TEXT
);
"""



class Defs(object):
    """General definitions that are valid for this survey.

    :ivar paper_width: Width of the paper in mm.
    :ivar paper_height: Height of the paper in mm.
    :ivar print_questionnaire_id: Whether a questionnaire ID is printed on each sheet.
    :ivar print_survey_id: Whether a survey ID is printed on each sheet.
    :ivar style: The style that is used for ID marking.
    :ivar duplex: Whether the questionnaire is duplex or not.
    """

    # Force a certain set of options using slots
    __slots__ = ['paper_width', 'paper_height', 'print_questionnaire_id',
                 'print_survey_id', 'style', 'duplex', 'checkmode',
                 'engine']

    def get_survey_id_pos(self):
        assert(self.style == 'classic')

        y_pos = self.paper_height - defs.corner_mark_bottom - defs.corner_box_padding
        y_pos -= defs.codebox_height

        left_padding = defs.corner_mark_left + 2 * defs.corner_box_padding + defs.corner_box_width
        right_padding = defs.corner_mark_right + 2 * defs.corner_box_padding + defs.corner_box_width

        text_y_pos = y_pos + defs.codebox_text_baseline_shift
        x_center = left_padding + (self.paper_width - left_padding - right_padding) / 2.0

        msb_box_x = left_padding
        lsb_box_x = self.paper_width - right_padding - defs.codebox_width

        text_x_pos = left_padding + (self.paper_width - right_padding - left_padding) / 2

        return msb_box_x, lsb_box_x, y_pos, text_x_pos, text_y_pos

    def get_questionnaire_id_pos(self):
        assert(self.style == 'classic')

        msb_box_x, lsb_box_x, y_pos, text_x_pos, text_y_pos = self.get_survey_id_pos()

        if self.print_survey_id:
            # Just move the y positions up if neccessary
            y_pos -= defs.codebox_height + defs.corner_box_padding
            text_y_pos -= defs.codebox_height + defs.corner_box_padding

        return msb_box_x, lsb_box_x, y_pos, text_x_pos, text_y_pos


class Survey(object):

    """The main survey object.

    :ivar defs: The :py:class:`model.survey.Defs` instance for this survey.
    :ivar survey_id: The survey ID of this survey.
    :ivar global_id: The global ID set for this survey. It is used during the "stamp" step.
    :ivar questionnaire: The py:class:`model.questionnaire.Questionnaire` instance representing the questionnaire.
    :ivar title: The title of the survey.
    :ivar info: Dictionary with general information about the survey.
    :ivar questionnaire_ids: A List of used questionnaire IDs.
    """

    _pickled_attrs = set(('defs', 'survey_id', 'questionnaire_ids', 'questionnaire', 'version'))

    def __init__(self):
        self.questionnaire = None

        self._internal_init()

        self.title = str()
        self.info = dict()
        self.survey_id = 0
        self.global_id = None
        self.questionnaire_ids = list()
        self.index = 0
        self.defs = Defs()
        self._db = None
        # Hardcoded for now.
        self._survey_rowid = 0

    def _internal_init(self):
        self._loaded_sheets = weakref.WeakValueDictionary()
        self._dirty_sheets = []
        self._current_sheet = None

    def add_questionnaire(self, questionnaire):
        self.questionnaire = questionnaire
        questionnaire.survey = self

    def add_sheet(self, sheet):
        """"
        WARNING: It is impossible to iterate newly added sheets before the
        survey has been saved!
        """
        sheet.survey = self
        sheet._rowid = -1

        self.goto_sheet(sheet)

    def delete_sheet(self, sheet):
        """"
        WARNING: The sheet will remain iteratable until the survey has been
        saved!
        """
        # This is a bit of a hack, but it will work great
        sheet._delete = True
        sheet._dirty = True

        self.goto_sheet(None)

    def calculate_survey_id(self):
        """Calculate the unique survey ID from the surveys settings and boxes.

        The ID only includes the boxes positions, which means that simple typo
        fixes will not change the ID most of the time."""
        import hashlib
        md5 = hashlib.new('md5')

        for qobject in self.questionnaire.qobjects:
            qobject.calculate_survey_id(md5)

        for defs_slot in self.defs.__slots__:
            # Backward compatibility
            if defs_slot == 'checkmode' and self.defs.checkmode == "checkcorrect":
                continue

            if defs_slot == 'engine':
                continue

            if isinstance(self.defs.__getattribute__(defs_slot), float):
                md5.update(str(round(self.defs.__getattribute__(defs_slot), 1)).encode('utf-8'))
            else:
                md5.update(str(self.defs.__getattribute__(defs_slot)).encode('utf-8'))

        # Use the first 32 bits as a little endian unsigned integer
        self.survey_id = struct.unpack('<I', md5.digest()[0:4])[0]

    @staticmethod
    def load(survey_dir):
        import configparser

        # Assume rowid is 0
        survey_rowid = 0

        dbfile = os.path.join(survey_dir, 'survey.sqlite')
        if not os.path.exists(dbfile):
            raise AssertionError('DB file does not exist!')
        _db = sqlite3.connect(dbfile)

        with _db as con:
            c = con.cursor()
            c.execute('SELECT json FROM surveys WHERE rowid=?', (survey_rowid,))
            _json, = c.fetchone()

            survey = db.fromJson(json.loads(_json), sys.modules[__name__])
            survey.survey_dir = survey_dir
            survey._survey_rowid = 0
            survey._db = _db

        ##########
        # Load the info file
        config = configparser.SafeConfigParser()
        config.optionxform = str
        config.read(os.path.join(survey_dir, 'info'))
        survey.title = config.get('sdaps', 'title')

        survey.global_id = config.get('sdaps', 'global_id')
        if survey.global_id == '' or survey.global_id == 'None':
            survey.global_id = None

        survey.info = dict()
        for key, value in config.items('info'):
            survey.info[key] = value

        return survey

    def _db_get_sheet(self, rowid):
        try:
            return self._loaded_sheets[rowid]
        except KeyError:
            pass

        c = self._db.cursor()
        c.execute('SELECT json FROM sheets WHERE survey_rowid=? AND rowid=?', (self._survey_rowid, rowid))

        data = json.loads(c.fetchone()[0])
        sheet = db.fromJson(data, Sheet)
        sheet._rowid = rowid
        sheet.survey = self
        sheet.reinit_state()

        self._loaded_sheets[rowid] = sheet
        return sheet

    def _db_save_sheet(self, cursor, sheet):
        if not sheet.dirty and sheet._rowid != -1:
            return

        if hasattr(sheet, '_delete') and sheet._delete:
            cursor.execute('DELETE FROM sheets WHERE survey_rowid=? and rowid=?', (self._survey_rowid, sheet._rowid))
            return

        tmp = json.dumps(sheet, default=db.toJson)
        if sheet._rowid == -1:
            cursor.execute('INSERT INTO sheets (survey_rowid, json) VALUES (?, ?)', (self._survey_rowid, tmp))
            sheet._rowid = cursor.lastrowid
            self._loaded_sheets[sheet._rowid] = sheet
        else:
            cursor.execute('UPDATE sheets SET json=? WHERE survey_rowid=? and rowid=?', (tmp, self._survey_rowid, sheet._rowid))

        sheet._clear_dirty()

    def __setstate__(self, state):
        self._internal_init()
        self.__dict__.update(state)
        self.questionnaire = db.fromJson(self.questionnaire, questionnaire.Questionnaire)
        self.questionnaire.survey = self
        self.defs = db.fromJson(self.defs, Defs)

        # Migrate old Defs object (from 1.9.5 or newer)
        if not hasattr(self.defs, 'engine'):
            self.defs.engine = defs.latex_engine

    @staticmethod
    def new(survey_dir):
        survey = Survey()
        survey.survey_dir = survey_dir
        try:
            os.makedirs(survey.path())
        except FileExistsError:
            pass

        dbfile = survey.path('survey.sqlite')
        if os.path.exists(dbfile):
            raise AssertionError('DB file already exists!')
        survey._db = sqlite3.connect(dbfile)
        survey._db.executescript(_db_schema)

        return survey

    def save(self):
        import configparser

        # Update the DB syncing out all changes
        with self._db as con:
            c = con.cursor()
            c.execute('INSERT OR REPLACE INTO surveys (rowid, json) VALUES (?, ?)', (self._survey_rowid, json.dumps(self, default=db.toJson)))

            for sheet in self._dirty_sheets:
                 self._db_save_sheet(c, sheet)
            if self._current_sheet:
                self._db_save_sheet(c, self._current_sheet)

            self._dirty_sheets = []

        # Sanity check that there is no sheet object alive that is dirty
        for sheet in self._loaded_sheets.values():
            assert sheet is None or (not sheet.dirty and sheet._rowid != -1)

        ###############
        # Write out the info file next

        # Hack to include comments. Set allow_no_value here, and add keys
        # with a '#' in the front and no value.
        config = configparser.SafeConfigParser(allow_no_value=True)

        config.optionxform = str
        config.add_section('sdaps')
        config.add_section('info')
        config.add_section('defs')
        config.add_section('questionnaire')
        config.set('sdaps', 'title', self.title)
        if self.global_id is not None:
            config.set('sdaps', 'global_id', self.global_id)
        else:
            config.set('sdaps', 'global_id', '')

        for key, value in sorted(self.info.items()):
            config.set('info', key, value)

        config.set('defs', '# These values are not read back, they exist for information only!')
        for attr in self.defs.__slots__:
            config.set('defs', attr, str(getattr(self.defs, attr)))

        config.set('questionnaire', '# These values are not read back, they exist for information only!')
        config.set('questionnaire', 'page_count', str(self.questionnaire.page_count))
        # Put the survey ID into "questionnaire". This seems sane even though
        # it is not stored there internally..
        config.set('questionnaire', 'survey_id', str(self.survey_id))

        # Atomically write info file
        info_fd = open(os.path.join(self.survey_dir, '.info.tmp'), 'w')
        config.write(info_fd)

        os.fsync(info_fd)
        info_fd.close()

        try:
            os.rename(os.path.join(self.survey_dir, 'info'), os.path.join(self.survey_dir, 'info~'))
        except OSError:
            pass

        os.rename(os.path.join(self.survey_dir, '.info.tmp'), os.path.join(self.survey_dir, 'info'))


    def path(self, *path):
        return os.path.join(self.survey_dir, *path)

    def new_path(self, path):
        content = os.listdir(self.path())
        i = 1
        while path % i in content:
            i += 1
        return os.path.join(self.survey_dir, path % i)

    def get_sheet(self):
        return self._current_sheet

    #: The currently selected sheet. Usually it will be changed by :py:meth:`iterate` or similar.
    sheet = property(get_sheet)

    def iterate(self, function, filter=lambda: True, *args, **kwargs):
        '''call function once for each sheet
        '''
        with self._db as con:
            c = con.cursor()
            c.execute('SELECT rowid FROM sheets WHERE survey_rowid=? ORDER BY sort,rowid', (self._survey_rowid,))
            for sheetid, in c.fetchall():
                self.goto_sheet(self._db_get_sheet(sheetid))
                if filter():
                    function(*args, **kwargs)

    @property
    def sheet_count(self):
        with self._db as con:
            c = con.cursor()
            c.execute('SELECT count(*) FROM sheets WHERE survey_rowid=?', (self._survey_rowid,))
            count = c.fetchone()[0]
            return count

    def iterate_progressbar(self, function, filter=lambda: True, *args, **kwargs):
        '''call function once for each sheet and display a progressbar
        '''
        with self._db as con:
            c = con.cursor()
            c.execute('SELECT count(*) FROM sheets WHERE survey_rowid=?', (self._survey_rowid,))
            count = c.fetchone()[0]

            # The old code used to first filter, and then run; but that is
            # a bit ineffective in a way
            print(ungettext('%i sheet', '%i sheets', count) % count)
            if count == 0:
                return

            log.progressbar.start(count)

            processed = 0
            c.execute('SELECT rowid FROM sheets WHERE survey_rowid=? ORDER BY sort,rowid', (self._survey_rowid,))
            for index, (sheetid,) in enumerate(c.fetchall()):
                self.goto_sheet(self._db_get_sheet(sheetid))
                if filter():
                    function(*args, **kwargs)
                    processed += 1

                log.progressbar.update(index + 1)

        print(_('Processed %i of %i sheets, took %f seconds') % (processed, count, log.progressbar.elapsed_time))

    def goto_sheet(self, sheet):
        '''goto the specified sheet object
        '''
        if self._current_sheet is not None and (self._current_sheet._rowid == -1 or self._current_sheet.dirty):
            if self._current_sheet not in self._dirty_sheets:
                self._dirty_sheets.append(self._current_sheet)

        self._current_sheet = sheet

    def goto_nth_sheet(self, index):
        with self._db as con:
            c = con.cursor()
            c.execute('SELECT rowid FROM sheets WHERE survey_rowid=? ORDER BY sort,rowid LIMIT 1 OFFSET ?', (self._survey_rowid, index))
            rowid = c.fetchone()[0]

        self.goto_sheet(self._db_get_sheet(rowid))

    def goto_questionnaire_id(self, questionnaire_id):
        '''goto the sheet object specified by its questionnaire_id
        '''

        qids = set()
        qids.add(questionnaire_id)
        # May also be an integer, so also add that if it can be decoded!
        try:
            qids.add(int(questionnaire_id))
        except ValueError:
            pass

        sheets = []
        def found():
            if self.sheet.questionnaire_id in qids:
                sheets.append(self.sheet)

        self.iterate(found)

        if len(sheets) == 1:
            self.goto_sheet(sheets[0])
        else:
            raise ValueError

    def check_settings(self):
        '''Do sanity checks on the different settings.'''

        if self.defs.duplex and self.questionnaire.page_count % 2 != 0:
            print(_("A questionnaire that is printed in duplex needs an even amount of pages!"))
            return False

        if self.defs.style == 'classic' and self.questionnaire.page_count > 6:
            print(_("The 'classic' style only supports a maximum of six pages! Use the 'code128' style if you require more pages."))
            return False

        return True

    def validate_questionnaire_id(self, qid):
        """Do style specific sanity checks on the questionnaire ID."""

        if self.defs.style == "classic":
            # The ID needs to be an integer
            try:
                return int(qid)
            except ValueError:
                log.error(_("IDs need to be integers in \"classic\" style!"))
                sys.exit(1)
        elif self.defs.style == "code128":
            # Check each character for validity
            for c in str(qid):
                if not c in defs.c128_chars:
                    log.error(_("Invalid character %s in questionnaire ID \"%s\" in \"code128\" style!") % (c, qid))
                    sys.exit(1)
            return qid
        elif self.defs.style == "custom":
            log.error(_("SDAPS cannot draw a questionnaire ID with the \"custom\" style. Do this yourself somehow!"))
            sys.exit(1)
        elif self.defs.style == "qr":
          return qid
        else:
            AssertionError()





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
The surface module adds support for loading the scanned images. It adds a buddy
to the model.sheet.Image and provides the surface via
model.sheet.Image.surface.surface at runtime.
"""

from . import model
from . import image


class Image(model.buddy.Buddy, metaclass=model.buddy.Register):
    """
    Buddy to load and cache image data. Do not forget to call :py:meth:`clean`
    when you are done (ie. before moving to the next sheet) or else the
    surface will be cached indefinately.

    :ivar surface: The cairo A1 surface, after it has been loaded using :py:meth:`load`.
    :ivar surface_rgb: The cairo RGB24 surface, after it has been loaded using :py:meth:`load_rgb`.
    """
    name = 'surface'
    obj_class = model.sheet.Image

    def load(self):
        """Load the A1 cairo surface, which is accessible using the surface
        attribute.
        :py:meth`clean` needs to be called when the surface is
        no longer needed."""
        self.surface = image.get_a1_from_tiff(
            self.obj.sheet.survey.path(self.obj.filename),
            self.obj.tiff_page,
            True if self.obj.rotated else False
        )

    def load_rgb(self):
        """Load the RGB24 cairo surface, which is accessible using the
        surface_rgb attribute.
        :py:meth:`clean` needs to be called when the surface is
        no longer needed."""
        self.surface_rgb = image.get_rgb24_from_tiff(
            self.obj.sheet.survey.path(self.obj.filename),
            self.obj.tiff_page,
            True if self.obj.rotated else False
        )

    def load_uncached(self):
        """Load the A1 surface and directly return it. It will not be cached,
        so using this function may result repeated reloads from file."""
        if hasattr(self, 'surface'):
            return self.surface
        else:
            return image.get_a1_from_tiff(
                self.obj.sheet.survey.path(self.obj.filename),
                self.obj.tiff_page,
                True if self.obj.rotated else False)

    def get_size(self):
        """Read the size of the surface. If the surface is already loaded, it
        will read the size from that. If it is not loaded, an uncached load will
        be done, which may be rather slow."""
        # Load uncached does not check the rgb surface
        if hasattr(self, 'surface_rgb'):
            s = self.surface_rgb
        else:
            s = self.load_uncached()
        return s.get_width(), s.get_height()

    def clean(self):
        """Call when you are done handling a specific sheet to free any cached
        cairo surface."""
        if hasattr(self, 'surface'):
            del self.surface
        if hasattr(self, 'surface_rgb'):
            del self.surface_rgb




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2010, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Contains the functionality to create a new SDAPS project using a LaTeX input
file.
"""

import sys
import os
import shutil
import glob

from sdaps.utils.mimetype import mimetype
from sdaps import model
from sdaps import log
from sdaps import paths
from sdaps import defs

from sdaps.utils import latex
from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

from ..setup import buddies
from . import sdapsfileparser
from ..setup import additionalparser


def setup(survey_dir, questionnaire_tex, engine, additionalqobjects=None, extra_files=[]):
    if os.access(survey_dir, os.F_OK):
        log.error(_('The survey directory already exists.'))
        return 1

    mime = mimetype(questionnaire_tex)
    if mime != 'text/x-tex' and mime != '':
        log.warn(_('Unknown file type (%s). questionnaire_tex should be of type text/x-tex.') % mime)
        log.warn(_('Will keep going, but expect failure!'))

    if additionalqobjects is not None:
        mime = mimetype(additionalqobjects)
        if mime != 'text/plain' and mime != '':
            log.error(_('Unknown file type (%s). additionalqobjects should be text/plain.') % mime)
            return 1

    # Create the survey directory, and copy the tex file.
    survey = model.survey.Survey.new(survey_dir)
    survey.defs.engine = engine

    # Add the new questionnaire
    survey.add_questionnaire(model.questionnaire.Questionnaire())

    try:
        shutil.copy(questionnaire_tex, survey.path('questionnaire.tex'))

        latex.write_override(survey, survey.path('sdaps.opt'), draft=True)

        # Copy class and dictionary files
        if paths.local_run:
            cls_extra_files = os.path.join(paths.source_dir, 'tex', '*.cls')
            cls_files = os.path.join(paths.source_dir, 'tex', 'class', 'build', 'local', '*.cls')
            tex_files = os.path.join(paths.source_dir, 'tex', 'class', 'build', 'local', '*.tex')
            sty_files = os.path.join(paths.source_dir, 'tex', 'class', 'build', 'local', '*.sty')
            dict_files = os.path.join(paths.build_dir, 'tex', '*.dict')
        else:
            cls_extra_files = None
            cls_files = os.path.join(paths.prefix, 'share', 'sdaps', 'tex', '*.cls')
            tex_files = os.path.join(paths.prefix, 'share', 'sdaps', 'tex', '*.tex')
            sty_files = os.path.join(paths.prefix, 'share', 'sdaps', 'tex', '*.sty')
            dict_files = os.path.join(paths.prefix, 'share', 'sdaps', 'tex', '*.dict')

        def copy_to_survey(files_glob):
            files = glob.glob(files_glob)
            for file in files:
                shutil.copyfile(file, survey.path(os.path.basename(file)))

        if cls_extra_files is not None:
            copy_to_survey(cls_extra_files)
        copy_to_survey(cls_files)
        copy_to_survey(tex_files)
        copy_to_survey(sty_files)
        copy_to_survey(dict_files)

        for add_file in extra_files:
            if os.path.isdir(add_file):
                shutil.copytree(add_file, survey.path(os.path.basename(add_file)))
            else:
                shutil.copyfile(add_file, survey.path(os.path.basename(add_file)))

        print(_("Running %s now multiple times to generate the questionnaire.") % survey.defs.engine)
        latex.compile(survey.defs.engine, 'questionnaire.tex', cwd=survey.path())

        if not os.path.exists(survey.path('questionnaire.pdf')):
            print(_("Error running \"%s\" to compile the LaTeX file.") % defs.latex_engine)
            raise AssertionError('PDF file not generated')

        survey.defs.print_questionnaire_id = False
        survey.defs.print_survey_id = True

        # Parse qobjects
        try:
            sdapsfileparser.parse(survey)

            for qobject in survey.questionnaire.qobjects:
                qobject.setup.setup()
                qobject.setup.validate()

        except:
            log.error(_("Caught an Exception while parsing the SDAPS file. The current state is:"))
            print(str(survey.questionnaire), file=sys.stderr)
            print("------------------------------------", file=sys.stderr)

            raise

        # Parse additionalqobjects
        if additionalqobjects:
            additionalparser.parse(survey, additionalqobjects)

        # Last but not least calculate the survey id
        survey.calculate_survey_id()

        if not survey.check_settings():
            log.error(_("Some combination of options and project properties do not work. Aborted Setup."))
            shutil.rmtree(survey.path())
            return 1

        # We need to now rebuild everything so that the correct ID is at the bottom
        # Dissable draft mode if the survey doesn't have questionnaire IDs
        latex.write_override(survey, survey.path('sdaps.opt'), draft=survey.defs.print_questionnaire_id)
        print(_("Running %s now multiple imes to generate the questionnaire.") % survey.defs.engine)
        os.remove(survey.path('questionnaire.pdf'))
        latex.compile(survey.defs.engine, 'questionnaire.tex', survey.path())

        if not os.path.exists(survey.path('questionnaire.pdf')):
            print(_("Error running \"%s\" to compile the LaTeX file.") % survey.defs.engine)
            raise AssertionError('PDF file not generated')

        # Print the result
        print(survey.title)

        for item in list(survey.info.items()):
            print('%s: %s' % item)

        log.logfile.open(survey.path('log'))

        survey.save()
        log.logfile.close()
    except:
        log.error(_("An error occured in the setup routine. The survey directory still exists. You can for example check the questionnaire.log file for LaTeX compile errors."))
        raise



# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2012, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module reorders already recognized data according to the questionnaire IDs.
"""

from collections import defaultdict
from sdaps import model

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

def reorder(survey):
    """We can assume quite some things in this function, because recognize
    properly handles it. ie.

     * Every image will be tagged correctly (as long as the data is known)
     * in duplex mode both the front/back image will be tagged, so we don't
       need to care about that here!
    """

    image_count = survey.questionnaire.page_count
    # We have two images per page in simplex mode!
    if not survey.defs.duplex:
        image_count = image_count * 2

    # First, go over all sheets and figure out which ones need reordering.
    # For every sheet that isn't quite right, we extract the images, and delete
    # the sheet.
    # The images are put into a dictionnary using the questionnaire ID.
    # Each entry in the dictionary is a list.
    images = defaultdict(lambda : [])

    # Load all sheets into memory
    sheets = []
    survey.iterate(lambda: sheets.append(survey.get_sheet()))

    for sheet in sheets: # Use a flat copy to iterate over
        broken = False
        pages = set()
        for image in sheet.images:
            if sheet.questionnaire_id != image.questionnaire_id:
                broken = True
            if sheet.global_id != image.global_id:
                broken = True

            # Check that no page exists twice
            if image.page_number is not None and image.page_number in pages:
                broken = True
            pages.add(image.page_number)

        # Also consider incomplete sets broken, so that hopefully they will
        # be filled up with the correct page.
        if len(sheet.images) != image_count:
            broken = True

        if broken:
            # Drop from the list of sheets
            survey.delete_sheet(sheet)

            for image in sheet.images:
                images[(image.questionnaire_id, image.global_id)].append(image)

    # We have dictionnary of lists of images that needs to be put into sheets
    # again.
    # This could probably be more robust. We don't care about the questionnaire
    # ID itself here, just put each list into sheets, splitting it into many
    # if there are too many images.
    for img_list in images.values():

        while len(img_list) > 0:
            sheet = model.sheet.Sheet()
            survey.add_sheet(sheet)

            while len(img_list) > 0 and len(sheet.images) < image_count:
                sheet.add_image(img_list.pop(0))


    survey.save()




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
sdaps has a modular design. It ships a core python module called "model" which
is responsible for storing and basic modification of the data. When other
modules like "recognize" are loaded they '''extend''' the original model.

As an example the "recognize" module contains everything required to analyize
the scanned data and find checkmarks. However, it will in addition load the
"image", "matrix" and "surface" modules. These three modules are responsible
for loading and caching the image data and doing some pre-processing of the
image data(transformation matrix calculation).

Please have a look at the documentation of the "model" package.
"""

import sys

from . import paths
from . import script
import os
import argparse

from .utils.ugettext import ugettext, ungettext
_ = ugettext


def init(local_run=False):
    paths.init(local_run, __path__[0])

def main(local_run=False):
    """The main SDAPS interface routine. It initilizes all modules, parses
    the command line and passes control over to the selected function."""
    init(local_run)

    from . import log
    log.activate_redirects()

    from . import cmdline

    cmdline = script.parser.parse_args()
    cmdline = vars(cmdline)

    log.interactive('-'*78 + '\n')
    log.interactive(('- SDAPS -- %s' % cmdline['_name']) + '\n')
    log.interactive('-'*78 + '\n')

    try:
        return cmdline['_func'](cmdline)
    except:
        import traceback
        string = traceback.format_exc()
        sys.stderr.write(string)
        sys.exit(1)



# Guess whether documentation is generated, if it is
# setup for local run.
if 'sphinx' in sys.argv[0]:
    paths.init(True, os.path.join(sys.path[0], 'sdaps'))




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2013, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys

from sdaps import model
from sdaps import script
from sdaps import log

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext


parser = script.add_project_subparser("convert",
    help=_("Convert a set of images to the correct image format."),
    description=_("""This command can be used if you scanned files in something
        other than the expected monochrome TIFF mode. All given files will
        be loaded, converted to monochrome and stored in a multipage 1bpp
        TIFF file. Optionally, you can select \"3D transformation"\ ,which may facilitate
        working with photos of questionnaires instead of scans."""))

parser.add_argument('--3d-transform',
    help=_("""Do a 3D-transformation after finding the corner marks.
    If they are not found, the image will be processed as-is."""),
    dest="transform",
    action="store_true",
    default=False)
parser.add_argument('-o', '--output', required=True,
    help=_("The location of the output file."),
    dest="output")

parser.add_argument('images',
    help=_("A number of TIFF image files."),
    nargs='+')

@script.connect(parser)
@script.logfile
def convert(cmdline):
    from sdaps.convert import convert_images

    if cmdline['output'] is None:
        log.error(_("No output filename specified!"))
        sys.exit(1)

    # We need a survey only for the paper size!
    survey = model.survey.Survey.load(cmdline['project'])

    convert_images(cmdline['images'], cmdline['output'], survey.defs.paper_width, survey.defs.paper_height, cmdline['transform'])





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2013, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from sdaps import image
from sdaps.utils import opencv
from sdaps import log

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

def convert_images(images, outfile, paper_width, paper_height, transform=False):

    portrait = paper_height >= paper_width

    for i, (img, filename, page) in enumerate(opencv.iter_images_and_pages(images)):
        img = opencv.ensure_orientation(img, portrait)
        img = opencv.sharpen(img)

        if transform:
            try:
                img = opencv.transform_using_corners(img, paper_width, paper_height)
            except AssertionError:
                log.warn(_("Could not apply 3D-transformation to image '%s', page %i!") % (filename, page))

        mono = opencv.convert_to_monochrome(img)
        image.write_a1_to_tiff(outfile, opencv.to_a1_surf(mono))





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2007-2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2007-2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Gdk
import cairo
import copy
import os
from sdaps import defs

from sdaps import model

from sdaps import matrix


class SheetWidget(Gtk.DrawingArea, Gtk.Scrollable):
    __gtype_name__ = "SDAPSSheetWidget"

    def __init__(self, provider):
        Gtk.DrawingArea.__init__(self)
        events = Gdk.EventMask.BUTTON_PRESS_MASK | Gdk.EventMask.BUTTON_RELEASE_MASK | \
            Gdk.EventMask.POINTER_MOTION_MASK | Gdk.EventMask.BUTTON_MOTION_MASK | \
            Gdk.EventMask.SCROLL_MASK | Gdk.EventMask.KEY_PRESS_MASK
        try:
            events = events | Gdk.EventMask.SMOOTH_SCROLL_MASK
        except AttributeError:
            # Does only work for GTK+ >=3.4
            pass
        self.add_events(events)

        self.hadj = None
        self.vadj = None
        self._hadj_value_changed_cb_id = None
        self._vadj_value_changed_cb_id = None

        self.provider = provider

        self._old_scroll_x = 0
        self._old_scroll_y = 0
        self._edge_drag_active = False

        self._zoom = 1.0

        self.props.can_focus = True

        self._update_matrices()
        self._cs_image = None
        self._ss_image = None

        self.provider.survey.questionnaire.connect_data_changed(self.partial_update)

    def update_state(self):
        # Cancel any dragging operation
        self._edge_drag_active = False
        self._update_matrices()
        self.queue_resize()
        self.queue_draw()

    def partial_update(self, questionnaire, qobj, obj, name, old_value):
        if not isinstance(obj, model.data.Box):
            return

        if self.provider.image.page_number != qobj.page_number:
            return

        self.invalidate_area(obj.x, obj.y, obj.width, obj.height)

    def _update_matrices(self):
        xoffset = 0
        yoffset = 0
        if self.hadj:
            xoffset = int(self.hadj.props.value)
        if self.vadj:
            yoffset = int(self.vadj.props.value)

        m = cairo.Matrix(self._zoom, 0,
                         0, self._zoom,
                         -xoffset, -yoffset)
        form_matrix = self.provider.image.matrix.mm_to_px()
        m = form_matrix.multiply(m)

        self._mm_to_widget_matrix = m
        self._widget_to_mm_matrix = \
            cairo.Matrix(*m)
        self._widget_to_mm_matrix.invert()

    def invalidate_area(self, x_mm, y_mm, width_mm, height_mm):
        x, y = self._mm_to_widget_matrix.transform_point(x_mm, y_mm)
        width, height = self._mm_to_widget_matrix.transform_distance(width_mm, height_mm)

        width = int(math.ceil(width + x - int(x))) + 20
        height = int(math.ceil(height + y - int(y))) + 20
        x = int(x) - 10
        y = int(y) - 10

        self.queue_draw_area(x, y, width, height)

    def _adjustment_changed_cb(self, adjustment):
        dx = int(self._old_scroll_x) - int(self.hadj.props.value)
        dy = int(self._old_scroll_y) - int(self.vadj.props.value)

        if self.get_window() is not None:
            self.get_window().scroll(dx, dy)

        self._old_scroll_x = self.hadj.props.value
        self._old_scroll_y = self.vadj.props.value

        # Update the transformation matrices
        self._update_matrices()

    def do_button_press_event(self, event):
        # Pass everything except normal clicks down
        if event.button != 1 and event.button != 2 and event.button != 3:
            return False

        if event.button == 2:
            self._drag_start_x = event.x
            self._drag_start_y = event.y
            cursor = Gdk.Cursor(Gdk.CursorType.HAND2)
            self.get_window().set_cursor(cursor)
            return True

        mm_x, mm_y = self._widget_to_mm_matrix.transform_point(event.x, event.y)

        if event.button == 3:
            # Give the corresponding widget the focus.
            box = self.provider.survey.questionnaire.gui.find_box(self.provider.image.page_number, mm_x, mm_y)
            if hasattr(box, "widget"):
                box.widget.focus()

            return True

        # button 1
        self.grab_focus()

        # Look for edges to drag first(on a 4x4px target)
        tollerance_x, tollerance_y = self._widget_to_mm_matrix.transform_distance(4.0, 4.0)
        result = self.provider.survey.questionnaire.gui.find_edge(self.provider.image.page_number, mm_x, mm_y,
                                                                  tollerance_x, tollerance_y)
        if result:
            self._edge_drag_active = True
            self._edge_drag_obj = result[0]
            self._edge_drag_data = result[1]
            return True

        box = self.provider.survey.questionnaire.gui.find_box(self.provider.image.page_number, mm_x, mm_y)

        if box is not None:
            box.data.state = not box.data.state

            return True

    def do_button_release_event(self, event):
        if event.button != 1 and event.button != 2 and event.button != 3:
            return False

        self.get_window().set_cursor(None)

        if event.button == 1:
            self._edge_drag_active = False

        return True

    def do_motion_notify_event(self, event):
        if event.state & Gdk.ModifierType.BUTTON2_MASK:
            x = int(event.x)
            y = int(event.y)

            dx = self._drag_start_x - x
            dy = self._drag_start_y - y

            if self.hadj:
                value = self.hadj.props.value + dx
                value = min(value, self.hadj.props.upper - self.hadj.props.page_size)
                self.hadj.set_value(value)
            if self.vadj:
                value = self.vadj.props.value + dy
                value = min(value, self.vadj.props.upper - self.vadj.props.page_size)
                self.vadj.set_value(value)

            self._drag_start_x = event.x
            self._drag_start_y = event.y

            return True
        elif event.state & Gdk.ModifierType.BUTTON1_MASK:
            if self._edge_drag_active:
                mm_x, mm_y = self._widget_to_mm_matrix.transform_point(event.x, event.y)

                self._edge_drag_obj.gui.move_edge(self._edge_drag_data, mm_x, mm_y)

                self.queue_draw()
                return True

        return False

    def do_get_request_mode(self):
        return Gtk.SizeRequestMode.CONSTANT_SIZE

    def do_get_preferred_height(self):
        if self.vadj:
            self.vadj.props.upper = self._render_height

        return min(self._render_height, 300), self._render_height

    def do_get_preferred_width(self):
        if self.hadj:
            self.hadj.props.upper = self._render_width

        return min(self._render_width, 300), self._render_width

    def do_size_allocate(self, allocation):
        # WTF? Why does this happen?
        if allocation.x < 0 or allocation.y < 0:
            GObject.idle_add(self.queue_resize)

        if self.hadj:
            self.hadj.props.page_size = min(self._render_width, allocation.width)
            if self.hadj.props.value > self._render_width - allocation.width:
                self.hadj.props.value = self._render_width - allocation.width
            self.hadj.props.page_increment = allocation.width * 0.9
            self.hadj.props.step_increment = allocation.width * 0.1
        if self.vadj:
            self.vadj.props.page_size = min(self._render_height, allocation.height)
            if self.vadj.props.value > self._render_height - allocation.height:
                self.vadj.props.value = self._render_height - allocation.height
            self.vadj.props.page_increment = allocation.height * 0.9
            self.vadj.props.step_increment = allocation.height * 0.1

        self._update_matrices()

        Gtk.DrawingArea.do_size_allocate(self, allocation)

    def do_draw(self, cr):

        cr.save()
        cr.save()

        # For the image
        xoffset = -int(self.hadj.props.value)
        yoffset = -int(self.vadj.props.value)

        image = self.provider.image.surface.surface_rgb

        if image != self._cs_image or self._ss_image is None:
            self._cs_image = image
            target = cr.get_target()
            self._ss_image = target.create_similar(cairo.CONTENT_COLOR, image.get_width(), image.get_height())
            subcr = cairo.Context(self._ss_image)
            subcr.set_source_surface(self._cs_image)
            subcr.paint()

        # Draw the image in the background
        cr.translate(xoffset, yoffset)
        cr.scale(self._zoom, self._zoom)

        cr.set_source_surface(self._ss_image, 0, 0)
        cr.paint()

        cr.restore()

        # Set the matrix _after_ drawing the background pixbuf.
        cr.transform(self._mm_to_widget_matrix)

        cr.set_source_rgba(1.0, 0.0, 0.0, 0.6)
        cr.set_line_width(1.0 * 25.4 / 72.0)
        cr.rectangle(defs.corner_mark_left, defs.corner_mark_top,
                     self.provider.survey.defs.paper_width - defs.corner_mark_left - defs.corner_mark_right,
                     self.provider.survey.defs.paper_height - defs.corner_mark_top - defs.corner_mark_bottom)
        cr.stroke()

        # Draw the overlay stuff.
        self.provider.survey.questionnaire.gui.draw(cr, self.provider.image.page_number)

        def inner_box(cr, x, y, width, height):
            line_width = cr.get_line_width()

            cr.rectangle(x + line_width / 2.0, y + line_width / 2.0,
                         width - line_width, height - line_width)

        if self.provider.survey.defs.style == 'classic':
            half_pt = 0.5 / 72.0 * 25.4
            pt = 1.0 / 72.0 * 25.4
            inner_box(cr,
                      defs.corner_mark_left + defs.corner_box_padding - half_pt,
                      defs.corner_mark_top + defs.corner_box_padding - half_pt,
                      defs.corner_box_width + pt,
                      defs.corner_box_height + pt)
            inner_box(cr,
                      self.provider.survey.defs.paper_width
                          - defs.corner_mark_right
                          - defs.corner_box_padding
                          - defs.corner_box_width - half_pt,
                      defs.corner_mark_top + defs.corner_box_padding - half_pt,
                      defs.corner_box_width + pt,
                      defs.corner_box_height + pt)
            inner_box(cr,
                      defs.corner_mark_left + defs.corner_box_padding - half_pt,
                      self.provider.survey.defs.paper_height
                          - defs.corner_mark_bottom
                          - defs.corner_box_padding
                          - defs.corner_box_height
                          - half_pt,
                      defs.corner_box_width + pt,
                      defs.corner_box_height + pt)
            inner_box(cr,
                      self.provider.survey.defs.paper_width
                          - defs.corner_mark_right
                          - defs.corner_box_padding
                          - defs.corner_box_width
                          - half_pt,
                      self.provider.survey.defs.paper_height
                          - defs.corner_mark_bottom
                          - defs.corner_box_padding
                          - defs.corner_box_height
                          - half_pt,
                      defs.corner_box_width + pt,
                      defs.corner_box_height + pt)
            cr.stroke()

        cr.restore()

        return True

    def do_key_press_event(self, event):
        if self.vadj:
            if event.keyval == Gdk.keyval_from_name("Up"):
                value = self.vadj.props.value - self.vadj.props.step_increment
                value = min(value, self.vadj.props.upper - self.vadj.props.page_size)
                self.vadj.set_value(value)
                return True
            if event.keyval == Gdk.keyval_from_name("Down"):
                value = self.vadj.props.value + self.vadj.props.step_increment
                value = min(value, self.vadj.props.upper - self.vadj.props.page_size)
                self.vadj.set_value(value)
                return True

        if self.hadj:
            if event.keyval == Gdk.keyval_from_name("Left"):
                value = self.hadj.props.value - self.hadj.props.step_increment
                value = min(value, self.hadj.props.upper - self.hadj.props.page_size)
                self.hadj.set_value(value)
                return True
            if event.keyval == Gdk.keyval_from_name("Right"):
                value = self.hadj.props.value + self.hadj.props.step_increment
                value = min(value, self.hadj.props.upper - self.hadj.props.page_size)
                self.hadj.set_value(value)
                return True
        return False

    def _get_render_width(self):
        image = self.provider.image.surface.surface_rgb
        if image:
            width = image.get_width()
        else:
            width = 400
        width = int(math.ceil(self._zoom * width))
        return width

    def _get_render_height(self):
        image = self.provider.image.surface.surface_rgb
        if image:
            height = image.get_height()
        else:
            height = 400

        height = int(math.ceil(self._zoom * height))
        return height

    _render_width = property(_get_render_width)
    _render_height = property(_get_render_height)

    def get_hscroll_policy(self):
        # Does not matter, we don't support natural sizes
        return Gtk.ScrollablePolicy.NATURAL
    hscroll_policy = \
        GObject.property(get_hscroll_policy,
                         type=Gtk.ScrollablePolicy,
                         default=Gtk.ScrollablePolicy.NATURAL)

    def get_vscroll_policy(self):
        # Does not matter, we don't support natural sizes
        return Gtk.ScrollablePolicy.NATURAL
    vscroll_policy = \
        GObject.property(get_vscroll_policy,
                         type=Gtk.ScrollablePolicy,
                         default=Gtk.ScrollablePolicy.NATURAL)

    def get_vadjustment(self):
        return self.vadj

    def set_vadjustment(self, value):
        if self._vadj_value_changed_cb_id is not None:
            self.vadj.disconnect(self._vadj_value_changed_cb_id)
            self._vadj_value_changed_cb_id = None
        self.vadj = value

        if self.vadj is not None:
            self._vadj_value_changed_cb_id = self.vadj.connect('value-changed', self._adjustment_changed_cb)

        self._update_matrices()

    vadjustment = GObject.property(get_vadjustment, set_vadjustment, type=Gtk.Adjustment)

    def get_hadjustment(self):
        return self.hadj

    def set_hadjustment(self, value):
        if self._hadj_value_changed_cb_id is not None:
            self.hadj.disconnect(self._hadj_value_changed_cb_id)
            self._hadj_value_changed_cb_id = None
        self.hadj = value

        if self.hadj is not None:
            self._hadj_value_changed_cb_id = self.hadj.connect('value-changed', self._adjustment_changed_cb)

        self._update_matrices()

    hadjustment = GObject.property(get_hadjustment, set_hadjustment, type=Gtk.Adjustment)

    def set_zoom(self, value):
        self._zoom = value
        self.queue_resize()

    def get_zoom(self):
        return self._zoom

    zoom = GObject.property(get_zoom, set_zoom, float, minimum=0.001, maximum=1024.0, default=1.0)





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2012, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from sdaps import model
from sdaps import script

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext


parser = script.add_project_subparser("reorder",
    help=_("Reorder pages according to questionnaire ID."),
    description=_("""This command reorders all pages according to the already
    recognized questionnaire ID. To use it add all the files to the project,
    then run a partial recognition using "recognize --identify". After this
    you have to run this command to reorder the data for the real recognition.
    """))


@script.connect(parser)
@script.logfile
def reorder(cmdline):
    survey = model.survey.Survey.load(cmdline['project'])

    from sdaps import reorder

    return reorder.reorder(survey)




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from sdaps import model
from sdaps import script

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

parser = script.add_project_subparser("report",
    help=_("Create a PDF report."),
    description=_("""This command creates a PDF report using reportlab that
    contains statistics and if selected the freeform fields."""))

parser.add_argument('-f', '--filter',
    help=_("Filter to only export a partial dataset."))
parser.add_argument('--all-filters',
    help=_("Create a filtered report for every checkbox."),
    action="store_true")
parser.add_argument('-s', '--short',
    help=_("Short format (without freeform text fields)."),
    action="store_const",
    const="short",
    dest="format")
parser.add_argument('-l', '--long',
    help=_("Detailed output. (default)"),
    dest="format",
    action="store_const",
    const="long",
    default="long")
parser.add_argument('--suppress-images',
    help=_('Do not include original images in the report. This is useful if there are privacy concerns.'),
    dest='suppress',
    action='store_const',
    const='images')
parser.add_argument('--suppress-substitutions',
    help=_('Do not use substitutions instead of images.'),
    dest='suppress',
    action='store_const',
    const='substitutions',
    default=None)
parser.add_argument('-p', '--paper',
    help=_('The paper size used for the output (default: locale dependent)'),
    dest='papersize')
parser.add_argument('-o', '--output',
    help=_("Filename to store the data to (default: report_%%i.pdf)"))

@script.connect(parser)
@script.logfile
def report(cmdline):
    from sdaps import report

    survey = model.survey.Survey.load(cmdline['project'])

    if cmdline['format'] == 'short':
        small = 1
    else:
        small = 0

    if cmdline['all_filters']:
        return report.stats(survey, cmdline['filter'], cmdline['output'], cmdline['papersize'], small, cmdline['suppress'])
    else:
        return report.report(survey, cmdline['filter'], cmdline['output'], cmdline['papersize'], small, cmdline['suppress'])





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo

from sdaps import model
from sdaps import surface
from sdaps import matrix


class Sheet(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'boxgallery'
    obj_class = model.sheet.Sheet

    def load(self):
        for image in self.obj.images:
            if not image.ignored:
                image.surface.load()

    def clean(self):
        for image in self.obj.images:
            if not image.ignored:
                image.surface.clean()


class Questionnaire(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'boxgallery'
    obj_class = model.questionnaire.Questionnaire

    def init(self, debugrecognition):
        self.checkboxes = list()
        self.debugrecognition = debugrecognition

    def clean(self):
        del self.checkboxes

    def get_checkbox_images(self):
        if self.obj.survey.sheet.valid:
            self.obj.survey.sheet.boxgallery.load()
            for qobject in self.obj.qobjects:
                self.checkboxes.extend(qobject.boxgallery.get_checkbox_images(self.debugrecognition))
            self.obj.survey.sheet.boxgallery.clean()


class QObject(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'boxgallery'
    obj_class = model.questionnaire.QObject

    def get_checkbox_images(self, debugrecognition):
        return []


class Question(QObject, metaclass=model.buddy.Register):

    name = 'boxgallery'
    obj_class = model.questionnaire.Question

    def get_checkbox_images(self, debugrecognition):
        boxes = []

        for box in self.obj.boxes:
            new_box = box.boxgallery.get_checkbox_image(debugrecognition)
            if new_box:
                boxes.append(new_box)
        return boxes


class Box(model.buddy.Buddy, metaclass=model.buddy.Register):
    name = 'boxgallery'
    obj_class = model.questionnaire.Box

    def get_checkbox_image(self, debugrecognition):
        return None


class Checkbox(model.buddy.Buddy, metaclass=model.buddy.Register):
    name = 'boxgallery'
    obj_class = model.questionnaire.Checkbox

    def get_checkbox_image(self, debugrecognition):
        image = self.obj.sheet.get_page_image(self.obj.page_number)

        border = 1.5
        mm_to_px = image.matrix.mm_to_px()

        px_x, px_y = mm_to_px.transform_point(
            self.obj.data.x - border, self.obj.data.y - border)
        px_x, px_y = int(px_x), int(px_y)
        px_width, px_height = mm_to_px.transform_distance(
            self.obj.data.width + 2 * border, self.obj.data.height + 2 * border)

        dest = cairo.ImageSurface(
            cairo.FORMAT_A1, int(px_width), int(px_height))
        src = image.surface.surface

        cr = cairo.Context(dest)
        cr.set_source_surface(src, -px_x, -px_y)
        cr.set_operator(cairo.OPERATOR_SOURCE)
        cr.paint()

        del cr
        dest.flush()

        debug = {}
        if debugrecognition:
            # Run the recognition for this checkbox
            self.obj.recognize.recognize()

            debug = self.obj.recognize.debug

        return (self.obj.data.state, self.obj.data.metrics, (dest, px_x, px_y), debug)






# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module contains buddy objects to calculate statistics from the data.

It is possible to search for large changes between different filters by calling
the "reference" function after a calculation. After this the significant boolean
will be set for new calculations (with different filters) if there is a large
deviation from the old value."""

import math

from sdaps import model


class Questionnaire(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'calculate'
    obj_class = model.questionnaire.Questionnaire

    def init(self):
        """Initialize or reset the state of the calculate module."""
        self.count = 0
        # iterate over qobjects
        for qobject in self.obj.qobjects:
            qobject.calculate.init()

    def read(self):
        """The function collects the data from a sheet. You should use
        :py:meth:`Survey.iterate` to call it for each sheet that needs to be
        counted."""
        self.count += 1
        # iterate over qobjects
        for qobject in self.obj.qobjects:
            qobject.calculate.read()

    def calculate(self):
        """Call once after :py:meth:`Questionnaire.calculate.read` to calculate
        statistical values like the standard deviation."""
        # iterate over qobjects
        for qobject in self.obj.qobjects:
            qobject.calculate.calculate()

    def reference(self):
        """Can be used to calculate a reference value. You can later check
        whether there was a significant difference to the previous run. The
        `significant` property will be set accordingly."""
        # iterate over qobjects
        for qobject in self.obj.qobjects:
            qobject.calculate.reference()


class QObject(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'calculate'
    obj_class = model.questionnaire.QObject

    def init(self):
        pass

    def read(self):
        pass

    def calculate(self):
        pass

    def reference(self):
        pass


class Question(QObject, metaclass=model.buddy.Register):

    name = 'calculate'
    obj_class = model.questionnaire.Question


class Choice(Question, metaclass=model.buddy.Register):
    """
    :ivar count: Number of times the question was answered.
    :ivar values: Dictionary for each box with the ratio the answer was choosen.
    :ivar significant: Whether there was a significant difference to the reference run.
    """
    name = 'calculate'
    obj_class = model.questionnaire.Choice

    def init(self):
        self.count = 0
        self.values = {box.value: 0 for box in self.obj.boxes}
        self.significant = {box.value: 0 for box in self.obj.boxes}

    def read(self):
        self.count += 1
        for item in self.obj.get_answer():
            self.values[item] += 1

    def calculate(self):
        if self.count:
            for value in self.values:
                self.values[value] = self.values[value] / float(self.count)
                if hasattr(self, 'ref_count'):
                    self.significant[value] = (
                        abs(self.values[value] - self.ref_values[value]) > 0.1)

    def reference(self):
        self.ref_count = self.count
        self.ref_values = self.values

class Option(Choice, metaclass=model.buddy.Register):

    name = 'calculate'
    obj_class = model.questionnaire.Option

    def read(self):
        answer = self.obj.get_answer()

        if answer is not False and answer >= 0:
            self.count += 1
            self.values[answer] += 1

class Range(Option, metaclass=model.buddy.Register):
    """
    :ivar count: Number of times the question was answered.
    :ivar values: Dictionary for each box with the ratio the value was choosen.
    :ivar mean: The average value that was choosen.
    :ivar standard_deviation: The average value that was choosen.
    :ivar significant: Whether there was a significant difference to the reference run.
    """
    name = 'calculate'
    obj_class = model.questionnaire.Range

    def init(self):
        Option.init(self)

        self.significant = 0
        self.mean = 0
        self.standard_deviation = 0
        self.range_count = 0

        self.range_values = {}
        for box in self.obj.boxes[self.obj.range[0]:self.obj.range[1] + 1]:
            self.range_values[box.value] = 0

        self.range_min = min(self.range_values)
        self.range_max = max(self.range_values)

    def calculate(self):
        self.mean = 0
        self.range_count = 0

        if self.count:
            for key in self.range_values:
                self.range_count += self.values[key]

                self.range_values[key] = self.values[key]
                del self.values[key]

            for key in self.values:
                self.values[key] = self.values[key] / float(self.count)

            if self.range_count > 0:
                # First calculate the mean
                for key in self.range_values:
                    self.mean += key * self.range_values[key]
                self.mean = self.mean / float(self.range_count)

                # Now we can calculate the standard deviation
                for key in self.range_values:
                    self.standard_deviation += self.range_values[key] * pow(key - self.mean, 2)
                self.standard_deviation = math.sqrt(self.standard_deviation / float(self.range_count))

                # And finally store the percentage rather than count for each answer
                for key in self.range_values:
                    self.range_values[key] = self.range_values[key] / float(self.count)

                if hasattr(self, 'ref_count'):
                    self.significant = abs(self.mean - self.ref_mean) > 0.1


    def reference(self):
        self.ref_count = self.count
        self.ref_mean = self.mean


class Additional_FilterHistogram(Question, metaclass=model.buddy.Register):

    name = 'calculate'
    obj_class = model.questionnaire.Additional_FilterHistogram

    def init(self):
        self.count = 0
        self.values = [0] * len(self.obj.answers)
        self.significant = [0] * len(self.obj.answers)

    def read(self):
        self.count += 1
        for i in range(len(self.obj.answers)):
            filter = clifilter.clifilter(
                self.obj.questionnaire.survey, self.obj.filters[i])
            if filter():
                self.values[i] += 1

    def calculate(self):
        if self.count:
            self.significant = dict()
            for i in range(len(self.values)):
                self.values[i] = self.values[i] / float(self.count)
                if hasattr(self, 'ref_count'):
                    self.significant[i] = (
                        abs(self.values[i] - self.ref_values[i]) > 0.1)

    def reference(self):
        self.ref_count = self.count
        self.ref_values = self.values





# Simply import all submodules

from . import add
from . import annotate
from . import boxgallery
from . import convert
from . import cover
from . import csvdata
from . import gui
from . import ids
from . import info
from . import recognize
from . import reorder
from . import report
from . import reporttex
from . import reset
from . import setup
from . import stamp





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import random

import os
import sys
import math
import codecs

from sdaps import model
from sdaps import log

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext


def stamp(survey, output_filename, cmdline):
    # copy questionnaire_ids
    # get number of sheets to create
    if cmdline['file'] or cmdline['random'] or cmdline['existing']:
        if not survey.defs.print_questionnaire_id:
            log.error(_("You may not specify the number of sheets for this survey. All questionnaires will be identical as the survey has been configured to not use questionnaire IDs for each sheet."))
            return 1

        if cmdline['existing']:
            questionnaire_ids = survey.questionnaire_ids
        elif cmdline['file']:
            if cmdline['file'] == '-':
                fd = sys.stdin
            else:
                fd = codecs.open(cmdline['file'], 'r', encoding="utf-8")

            questionnaire_ids = list()
            for line in fd.readlines():
                # Only strip newline/linefeed not spaces
                line = line.strip('\n\r')

                # Skip empty lines
                if line == "":
                    continue

                questionnaire_ids.append(survey.validate_questionnaire_id(line))
        else:
            # Create random IDs
            max = pow(2, 16)
            min = max - 50000
            questionnaire_ids = list(range(min, max))

            # Remove any id that has already been used.
            for id in survey.questionnaire_ids:
                if type(id) != int:
                    continue
                questionnaire_ids[id - min] = 0

            questionnaire_ids = [id for id in questionnaire_ids if id > min]
            random.shuffle(questionnaire_ids)
            questionnaire_ids = questionnaire_ids[:cmdline['random']]
    else:
        if survey.defs.print_questionnaire_id:
            log.error(_("This survey has been configured to use questionnaire IDs. Each questionnaire will be unique. You need to use on of the options to add new IDs or use the existing ones."))
            return 1

        questionnaire_ids = None

    if questionnaire_ids is not None and not cmdline['existing']:
        survey.questionnaire_ids.extend(questionnaire_ids)

    if os.path.exists(survey.path('questionnaire.tex')):
        # use the LaTeX stamper
        from sdaps.stamp.latex import create_stamp_pdf
    else:
        raise AssertionError('Only LaTeX stamping is currently supported!')

    create_stamp_pdf(survey, output_filename, questionnaire_ids)

    survey.save()




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from sdaps import model


def parse(survey, additionalqobjects):

    document = file(additionalqobjects, 'r')

    for line in document:
        line = line.decode('utf-8')
        args = line.strip().split('\t')
        qobject = getattr(model.questionnaire, 'Additional_%s' % args.pop(0))
        assert issubclass(qobject, model.questionnaire.QObject)
        qobject = qobject()
        survey.questionnaire.add_qobject(qobject)
        qobject.setup.setup(args)




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008,2013, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Example of using this to preload your own custom buddy. This first sets up the
# paths, below the call into sdaps happens
if __name__ == '__main__':
    import sys
    import os
    sys.path.append(os.path.join(os.path.dirname(__file__), '..', '..'))

from sdaps import defs

from sdaps import model
from sdaps.utils.exceptions import RecognitionError

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext


class Image(model.buddy.Buddy, metaclass=model.buddy.Register):
    """Example buddy class for a "blank" style. It simply assumes
    that everything is good about the page. ie. first page, no rotation,
    correct survey ID, and no other IDs."""
    name = 'style'
    obj_class = model.sheet.Image

    def get_page_rotation(self):
        """Return the rotation of the passed image."""
        # Assume not rotated
        return False

    def get_page_number(self):
        """Return page number or None if not known."""
        # Assume only one page
        return 1

    def get_survey_id(self):
        """Return the survey ID as read from the image, or None."""
        # Return the expected survey ID
        return self.obj.sheet.survey.survey_id

    def get_questionnaire_id(self):
        """Return the questionnaire ID as read from the image."""
        # Assume no (useful) questionnaire ID exists
        return None

    def get_global_id(self):
        """Return the global ID as read from the image."""
        # Assume no (useful) global ID exists
        return None

# Example of using this to preload your own custom buddy. This assume a local run.
# You can of course use the API directly too.
if __name__ == '__main__':
    import sdaps
    sys.exit(sdaps.main(local_run = True))




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math

from reportlab import platypus
from reportlab.lib import styles
from reportlab.lib import colors
from reportlab.lib import units

from sdaps import clifilter
from sdaps import template
from sdaps import model

from xml.sax.saxutils import escape

from . import flowables
from . import answers

from sdaps import calculate


mm = units.mm


stylesheet = dict(template.stylesheet)

stylesheet['Head'] = styles.ParagraphStyle(
    'Head',
    stylesheet['Normal'],
    fontSize=12,
    leading=17,
    backColor=colors.lightgrey,
    spaceBefore=5 * mm,
)

stylesheet['Question'] = styles.ParagraphStyle(
    'Question',
    stylesheet['Normal'],
    spaceBefore=3 * mm,
    fontName='Times-Bold',
)

stylesheet['Text'] = styles.ParagraphStyle(
    'Text',
    stylesheet['Normal'],
    spaceBefore=1 * mm,
    spaceAfter=1 * mm,
    rightIndent=5 * mm,
    bulletIndent=2 * mm,
    leftIndent=5 * mm,
)


class Questionnaire(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'report'
    obj_class = model.questionnaire.Questionnaire

    def init(self, small=0, suppress=None):
        self.small = small
        # iterate over qobjects
        for qobject in self.obj.qobjects:
            qobject.report.init(small, suppress)

    def report(self):
        # iterate over qobjects
        for qobject in self.obj.qobjects:
            qobject.report.report()

    def story(self):
        story = list()
        # iterate over qobjects
        keeptogether_list = []
        for qobject in self.obj.qobjects:
            new, keeptogether = qobject.report.story()
            new = list(new)

            if len(new) == 0:
                continue

            if keeptogether:
                keeptogether_list.extend(new)
            else:
                if len(keeptogether_list):
                    add = new.pop(0)
                    if isinstance(add, platypus.KeepTogether):
                        keeptogether_list.extend(add._content)
                    else:
                        keeptogether_list.append(add)

                    story.append(platypus.KeepTogether(keeptogether_list))
                    keeptogether_list = []

                story.extend(new)

        story.extend(keeptogether_list)
        return story

    def filters(self):
        filters = list()
        # iterate over qobjects
        for qobject in self.obj.qobjects:
            filters.extend(list(qobject.report.filters()))
        return filters


class QObject(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'report'
    obj_class = model.questionnaire.QObject

    def init(self, small, suppress):
        self.small = small

    def report(self):
        pass

    def story(self):
        return [], False

    def filters(self):
        return []


class Head(QObject, metaclass=model.buddy.Register):

    name = 'report'
    obj_class = model.questionnaire.Head

    def story(self):
        return [
            platypus.Paragraph(
                '%s %s' % (self.obj.id_str(), escape(self.obj.title)),
                stylesheet['Head'])], True


class Question(QObject, metaclass=model.buddy.Register):

    name = 'report'
    obj_class = model.questionnaire.Question

    def story(self):
        return [
            platypus.Paragraph(
                '%s %s' % (
                    self.obj.id_str(), escape(self.obj.question)),
                stylesheet['Question'])], True


class Choice(Question, metaclass=model.buddy.Register):

    name = 'report'
    obj_class = model.questionnaire.Choice

    def init(self, small, suppress):
        self.small = small
        self.suppress = suppress
        self.text = list()

    def report(self):
        if not self.small:
            for box in self.obj.boxes:
                if (isinstance(box, model.questionnaire.Textbox) and
                        box.data.state):

                    if box.data.text and self.suppress != 'substitutions':
                        self.text.append(answers.RawText(box.data.text,
                                                         stylesheet['Text']))
                    elif self.suppress != 'images':
                        self.text.append(answers.Freeform(box))

    def story(self):
        story, tmp = Question.story(self)
        if self.obj.calculate.count:
            for box in self.obj.boxes:
                story.append(
                    answers.Choice(
                        box.text,
                        self.obj.calculate.values[box.value],
                        self.obj.calculate.significant[box.value]
                    )
                )
            story = [platypus.KeepTogether(story)]
            if len(self.text) > 0:
                story.append(platypus.Spacer(0, 3 * mm))
                story.extend(self.text)
        return story, False

    def filters(self):
        for box in self.obj.boxes:
            yield '%i in %s' % (box.value, self.obj.id_filter())

class Option(Question, metaclass=model.buddy.Register):

    name = 'report'
    obj_class = model.questionnaire.Option

    def filters(self):
        for box in self.obj.boxes:
            yield '%i == %s' % (box.value, self.obj.id_filter())

class Range(Option, metaclass=model.buddy.Register):

    name = 'report'
    obj_class = model.questionnaire.Range

    def report(self):
        pass

    def story(self):
        story, tmp = Question.story(self)
        if self.obj.calculate.range_count:
            story.append(answers.Range(
                self.obj.calculate.range_min,
                self.obj.calculate.range_max,
                self.obj.calculate.range_values,
                self.obj.answers,
                self.obj.calculate.mean,
                self.obj.calculate.standard_deviation,
                self.obj.calculate.count,
                self.obj.calculate.significant))

        if self.obj.calculate.count > 0 and self.obj.calculate.values:
            for box in self.obj.boxes:
                if box.value in self.obj.calculate.values:
                    story.append(
                        answers.Choice(
                            box.text,
                            self.obj.calculate.values[box.value],
                            False
                        )
                    )

        story = [platypus.KeepTogether(story)]
        return story, False


class Text(Question, metaclass=model.buddy.Register):

    name = 'report'
    obj_class = model.questionnaire.Text

    def init(self, small, suppress):
        self.small = small
        self.suppress = suppress
        self.text = list()

    def report(self):
        if not self.small:
            for box in self.obj.boxes:
                if box.data.state:
                    if box.data.text and self.suppress != 'substitutions':
                        self.text.append(answers.RawText(box.data.text,
                                                         stylesheet['Text']))
                    elif self.suppress != 'images':
                        self.text.append(answers.Freeform(box))

    def story(self):
        story, tmp = Question.story(self)
        if len(self.text) > 0:
            story.append(self.text[0])
            story = [platypus.KeepTogether(story)]
        if len(self.text) > 1:
            story.extend(self.text[1:])
        return story, False


class Additional_FilterHistogram(Question, metaclass=model.buddy.Register):

    name = 'report'
    obj_class = model.questionnaire.Additional_FilterHistogram

    def story(self):
        story, tmp = Question.story(self)
        if self.obj.calculate.count:
            for i in range(len(self.obj.calculate.values)):
                story.append(
                    answers.Choice(
                        self.obj.answers[i],
                        self.obj.calculate.values[i],
                        self.obj.calculate.significant[i]
                    )
                )
            story = [platypus.KeepTogether(story)]
        return story, False



# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, 2011, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo
import math
import os

from sdaps import clifilter
from sdaps import template
from sdaps import model
from sdaps import image

from sdaps import calculate

from sdaps.utils.latex import raw_unicode_to_latex, unicode_to_latex

from sdaps.utils.image import ImageWriter


def format_raw_text(text):
    from sdaps.setuptex import latexmap

class Questionnaire(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'report'
    obj_class = model.questionnaire.Questionnaire

    def init(self, img_dir, small=0, suppress=None):
        self.small = small

        self.textbox_writer = ImageWriter(img_dir, 'textbox-')

        # iterate over qobjects
        for qobject in self.obj.qobjects:
            qobject.report.init(small, suppress)

    def report(self, tmpdir):
        # iterate over qobjects
        for qobject in self.obj.qobjects:
            qobject.report.report(tmpdir)

    def write(self, out, tmpdir):
        # iterate over qobjects
        for qobject in self.obj.qobjects:
            qobject.report.write(out, tmpdir)

    def filters(self):
        filters = list()
        # iterate over qobjects
        for qobject in self.obj.qobjects:
            filters.extend(list(qobject.report.filters()))
        return filters


class QObject(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'report'
    obj_class = model.questionnaire.QObject

    def init(self, small, suppress):
        self.small = small

    def report(self, tmpdir):
        pass

    def write(self, out, tmpdir):
        pass

    def filters(self):
        return []


class Head(QObject, metaclass=model.buddy.Register):

    name = 'report'
    obj_class = model.questionnaire.Head

    def write(self, out, tmpdir):
        # Smarter numbering handling?
        out.write('\\section*{%s %s}\n' % (self.obj.id_str(), unicode_to_latex(self.obj.title)))


class Question(QObject, metaclass=model.buddy.Register):

    name = 'report'
    obj_class = model.questionnaire.Question

    def write_begin(self, out):
        # Smarter numbering handling?
        out.write('\\begin{question}{%s %s}\n' % (self.obj.id_str(), unicode_to_latex(self.obj.question)))

    def write_end(self, out):
        # Smarter numbering handling?
        out.write('\\end{question}\n\n')

    def write(self, out, tmpdir):
        self.write_begin(out)
        self.write_end(out)


class Choice(Question, metaclass=model.buddy.Register):

    name = 'report'
    obj_class = model.questionnaire.Choice

    def init(self, small, suppress):
        self.small = small
        self.suppress = suppress
        self.text = ""

    def report(self, tmpdir):
        if not self.small:
            for box in self.obj.boxes:
                if (isinstance(box, model.questionnaire.Textbox) and
                        box.data.state):
                    if box.data.text and self.suppress != 'substitutions':
                        text = raw_unicode_to_latex(box.data.text)
                        self.text += '\\freeformtext{%s}\n' % (text)
                    elif self.suppress != 'images':
                        img_file = self.obj.questionnaire.report.textbox_writer.output_box(box)
                        self.text += '\\freeform{%fmm}{%s}\n' % (box.data.width, img_file)

    def write_begin(self, out):
        # Smarter numbering handling?
        out.write('\\begin{choicequestion}{%s %s}\n' % (self.obj.id_str(), unicode_to_latex(self.obj.question)))

    def write_end(self, out):
        # Smarter numbering handling?
        out.write('\\end{choicequestion}\n\n')

    def write(self, out, tmpdir):
        self.write_begin(out)
        if self.obj.calculate.count:
            for box in self.obj.boxes:
                out.write('''\\choiceanswer{%s}{%.3f}\n''' % (unicode_to_latex(box.text), self.obj.calculate.values[box.value]))
        self.write_end(out)

        out.write(self.text)

    def filters(self):
        for box in self.obj.boxes:
            yield '%i in %s' % (box.value, self.obj.id_filter())

class Option(Choice, metaclass=model.buddy.Register):

    name = 'report'
    obj_class = model.questionnaire.Option

    def filters(self):
        for box in self.obj.boxes:
            yield '%i == %s' % (box.value, self.obj.id_filter())

class Range(Question, metaclass=model.buddy.Register):

    name = 'report'
    obj_class = model.questionnaire.Range

    def write(self, out, tmpdir):
        Question.write_begin(self, out)

        if self.obj.calculate.range_count:
            out.write('\\pgfkeyssetvalue{/sdaps/mark/range}{%s}\n' % (self.obj.calculate.range_max - self.obj.calculate.range_min + 1))
            out.write('\\pgfkeyssetvalue{/sdaps/mark/rangemin}{%s}\n' % (self.obj.calculate.range_min))
            out.write('\\pgfkeyssetvalue{/sdaps/mark/rangemax}{%s}\n' % (self.obj.calculate.range_max))
            out.write('\\pgfkeyssetvalue{/sdaps/mark/lower}{%s}\n' % (unicode_to_latex(self.obj.answers[0])))
            out.write('\\pgfkeyssetvalue{/sdaps/mark/upper}{%s}\n' % (unicode_to_latex(self.obj.answers[1])))
            out.write('\\pgfkeyssetvalue{/sdaps/mark/count}{%i}\n' % (self.obj.calculate.count))
            out.write('\\pgfkeyssetvalue{/sdaps/mark/stddev}{%.1f}\n' % (self.obj.calculate.standard_deviation))
            for key in range(self.obj.calculate.range_min, self.obj.calculate.range_max + 1):
                if key in self.obj.calculate.range_values:
                    fraction = self.obj.calculate.range_values[key]
                else:
                    fraction = 0
                out.write('\\pgfkeyssetvalue{/sdaps/mark/%i/fraction}{%.3f}\n' % (key, fraction))
            out.write('\\pgfkeyssetvalue{/sdaps/mark/mean}{%.1f}\n' % (self.obj.calculate.mean))
            out.write('\n\\markanswer\n')

        if self.obj.calculate.count > 0 and self.obj.calculate.values:
            out.write('\\begin{embedchoicequestion}\n')
            for box in self.obj.boxes:
                if box.value in self.obj.calculate.values:
                    out.write('''\\choiceanswer{%s}{%.3f}\n''' % (unicode_to_latex(box.text), self.obj.calculate.values[box.value]))
            out.write('\\end{embedchoicequestion}\n')

        Question.write_end(self, out)

    def filters(self):
        for x in range(len(self.obj.boxes)+1):
            yield '%i == %s' % (x, self.obj.id_filter())


class Text(Question, metaclass=model.buddy.Register):

    name = 'report'
    obj_class = model.questionnaire.Text

    def init(self, small, suppress):
        self.small = small
        self.suppress = suppress
        self.text = ""

    def report(self, tmpdir):
        if not self.small:
            for box in self.obj.boxes:
                if box.data.state:
                    if box.data.text and self.suppress != 'substitutions':
                        text = raw_unicode_to_latex(box.data.text)
                        self.text += '\\freeformtext{%s}\n' % (text)
                    elif self.suppress != 'images':
                        img_file = self.obj.questionnaire.report.textbox_writer.output_box(box)
                        self.text += '\\freeform{%fmm}{%s}\n' % (box.data.width, img_file)

    def write(self, out, tmpdir):
        Question.write_begin(self, out)

        out.write(self.text)

        Question.write_end(self, out)


class Additional_FilterHistogram(Question, metaclass=model.buddy.Register):

    name = 'report'
    obj_class = model.questionnaire.Additional_FilterHistogram

    def write(self, tmpdir):
        Question.write_begin(self, out)

        if self.obj.calculate.count:
            for i in range(len(self.obj.calculate.values)):
                out.write('''\\choiceanswer{%s}{%.3f}{%.3f}\n''' %
                          (unicode_to_latex(self.obj.answers[i]), self.obj.calculate.values[i], self.obj.calculate.significant[i]))

        Question.write_end(self, out)




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008,2011, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo
import math

from sdaps import model
from sdaps import matrix
from sdaps import surface
from sdaps import image
from sdaps import defs
from sdaps.utils.exceptions import RecognitionError
from sdaps import log

from sdaps.utils.barcode import read_barcode
from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

pt_to_mm = 25.4 / 72.0

warned_multipage_not_correctly_scanned = False


class Sheet(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'recognize'
    obj_class = model.sheet.Sheet

    def recognize(self):
        global warned_multipage_not_correctly_scanned

        self.obj.valid = 1

        duplex_mode = self.obj.survey.defs.duplex

        # Load all images of this sheet
        for image in self.obj.images:
            if not image.ignored:
                image.rotated = 0
                image.surface.load()

        failed_pages = set()

        # Matrix recognition for all of them
        matrix_errors = set()
        for page, image in enumerate(self.obj.images):
            try:
                image.recognize.calculate_matrix()
            except RecognitionError:
                matrix_errors.add(page)

        # We need to check the matrix_errors. Some are expected in simplex mode
        for page in matrix_errors:
            # in simplex mode every page will have a matrix; it might be a None
            # matrix though

            log.warn(_('%s, %i: Matrix not recognized.') % (self.obj.images[page].filename, self.obj.images[page].tiff_page))
            failed_pages.add(page)

        # Rotation for all of them
        for page, image in enumerate(self.obj.images):
            try:
                # This may set the rotation to "None" for unknown
                image.recognize.calculate_rotation()
            except RecognitionError:
                log.warn(_('%s, %i: Rotation not found.') % (image.filename, image.tiff_page))
                failed_pages.add(page)

        # Copy the rotation over (if required) and print warning if the rotation is unknown
        self.duplex_copy_image_attr(failed_pages, 'rotated', _("Neither %s, %i or %s, %i has a known rotation!"))

        # Reload any image that is rotated.
        for page, image in enumerate(self.obj.images):
            if image.rotated and not image.ignored:
                image.surface.load()
                # And redo the whole matrix stuff ...
                # XXX: It would be better to manipulate the matrix instead.
                try:
                    image.recognize.calculate_matrix()
                except RecognitionError:
                    if duplex_mode:
                        log.warn(_('%s, %i: Matrix not recognized (again).') % (image.filename, image.tiff_page))
                        failed_pages.add(page)

        ############
        # At this point we can extract the page numbers and IDs as neccessary.
        ############

        # Figure out the page numbers
        # ***************************
        for page, image in enumerate(self.obj.images):
            try:
                # This may set the page_number to "None" for unknown
                image.recognize.calculate_page_number()
            except RecognitionError:
                log.warn(_('%s, %i: Could not get page number.') % (image.filename, image.tiff_page))
                image.page_number = None
                failed_pages.add(page)

        i = 0
        while i < len(self.obj.images):
            # We try to recover at least the page number of failed pages
            # this way.
            # NOTE: In simplex mode dummy pages will be inserted, so one page
            # always has no page number, and the other one has one.
            # This is exactly what we want, so we don't need to do anything
            # (except warn if we did not find any page!)
            failed = (i in failed_pages or i + 1 in failed_pages)

            first = self.obj.images[i]
            second = self.obj.images[i + 1]

            if first.page_number is None and second.page_number is None:
                if not failed:
                    # Whoa, that should not happen.
                    log.warn(_("Neither %s, %i or %s, %i has a known page number!" %
                             (first.filename, first.tiff_page, second.filename, second.tiff_page)))
                    failed_pages.add(i)
                    failed_pages.add(i + 1)

            elif duplex_mode == False:
                # Simplex mode is special, as we know that one has to be unreadable
                # we need to ensure one of the page numbers is None
                if first.page_number is not None and second.page_number is not None:
                    # We don't touch the ignore flag in this case
                    # Simply print a message as this should *never* happen
                    log.error(_("Got a simplex document where two adjacent pages had a known page number. This should never happen as even simplex scans are converted to duplex by inserting dummy pages. Maybe you did a simplex scan but added it in duplex mode? The pages in question are %s, %i and %s, %i.") % (first.filename, first.tiff_page, second.filename, second.tiff_page))

                # Set the ignored flag for the unreadable page. This is a valid
                # operation as the back side of a readable page is known to be
                # empty.
                elif first.page_number is None:
                    first.ignored = True
                else:
                    second.ignored = True

            elif first.page_number is None:
                # One based, odd -> +1, even -> -1
                first.page_number = second.page_number - 1 + 2 * (second.page_number % 2)
            elif second.page_number is None:
                second.page_number = first.page_number - 1 + 2 * (first.page_number % 2)
            elif first.page_number != (second.page_number - 1 + 2 * (second.page_number % 2)):
                if not failed:
                    log.warn(_("Images %s, %i and %s, %i do not have consecutive page numbers!" %
                             (first.filename, first.tiff_page, second.filename, second.tiff_page)))

                    failed_pages.add(i)
                    failed_pages.add(i + 1)

            i += 2

        # Check that every page has a non None value, and each page exists once.
        pages = set()
        for i, image in enumerate(self.obj.images):
            # Ignore known blank pages
            if image.ignored:
                continue

            if image.page_number is None:
                log.warn(_("No page number for page %s, %i exists." % (image.filename, image.tiff_page)))
                failed_pages.add(i)
                continue

            if image.page_number in pages:
                log.warn(_("Page number for page %s, %i already used by another image.") %
                         (image.filename, image.tiff_page))
                failed_pages.add(i)
                continue

            if image.page_number <= 0 or image.page_number > self.obj.survey.questionnaire.page_count:
                log.warn(_("Page number %i for page %s, %i is out of range.") %
                         (image.page_number, image.filename, image.tiff_page))
                failed_pages.add(i)
                continue

            pages.add(image.page_number)

        # Figure out the suvey ID if neccessary
        # *************************************
        if self.obj.survey.defs.print_survey_id:
            for page, image in enumerate(self.obj.images):
                try:
                    if not duplex_mode or (image.page_number is not None and image.page_number % 2 == 0):
                        image.recognize.calculate_survey_id()
                    else:
                        image.survey_id = None
                except RecognitionError:
                    log.warn(_('%s, %i: Could not read survey ID, but should be able to.') %
                             (image.filename, image.tiff_page))
                    failed_pages.add(page)

            self.duplex_copy_image_attr(failed_pages, "survey_id", _("Could not read survey ID of either %s, %i or %s, %i!"))

            # Simply use the survey ID from the first image globally
            self.obj.survey_id = self.obj.images[0].survey_id

            if self.obj.survey_id != self.obj.survey.survey_id:
                # Broken survey ID ...
                log.warn(_("Got a wrong survey ID (%s, %i)! It is %s, but should be %i.") %
                         (self.obj.images[0].filename,
                          self.obj.images[0].tiff_page,
                          self.obj.survey_id,
                          self.obj.survey.survey_id))
                self.obj.valid = 0
        else:
            # Assume that the data is from the correct survey
            self.obj.survey_id = self.obj.survey.survey_id
            for image in self.obj.images:
                image.survey_id = self.obj.survey.survey_id

        # Figure out the questionnaire ID if neccessary
        # *********************************************
        if self.obj.survey.defs.print_questionnaire_id:
            questionnaire_ids = []

            for page, image in enumerate(self.obj.images):
                try:
                    if not duplex_mode or (image.page_number is not None and image.page_number % 2 == 0):
                        image.recognize.calculate_questionnaire_id()
                except RecognitionError:
                    log.warn(_('%s, %i: Could not read questionnaire ID, but should be able to.') % \
                             (image.filename, image.tiff_page))
                    failed_pages.add(page)
                if image.questionnaire_id is not None:
                    questionnaire_ids.append(image.questionnaire_id)

            self.duplex_copy_image_attr(failed_pages, "questionnaire_id", _("Could not read questionnaire ID of either %s, %i or %s, %i!"))

            if len(questionnaire_ids):
                self.obj.questionnaire_id = questionnaire_ids[0]
            else:
                self.obj.questionnaire_id = self.obj.images[0].questionnaire_id

        # Try to load the global ID. If it does not exist we will get None, if
        # it does, then it will be non-None. We don't care much about it
        # internally anyways.
        # However, we do want to ensure that it is the same everywhere if it
        # can be read in.
        # *********************************************
        for page, image in enumerate(self.obj.images):
            try:
                if not duplex_mode or (image.page_number is not None and image.page_number % 2 == 0):
                    image.recognize.calculate_global_id()
            except RecognitionError:
                pass

        self.duplex_copy_image_attr(failed_pages, "global_id")

        self.obj.global_id = self.obj.images[0].global_id

        for image in self.obj.images:
            if self.obj.global_id != image.global_id or \
                self.obj.survey_id != image.survey_id or \
                self.obj.questionnaire_id != image.questionnaire_id:

                if not warned_multipage_not_correctly_scanned:
                    log.warn(_("Got different IDs on different pages for at least one sheet! Do *NOT* try to use filters with this survey! You have to run a \"reorder\" step for this to work properly!"))

                    warned_multipage_not_correctly_scanned = True

        # Done
        if failed_pages:
            self.obj.valid = 0

    def clean(self):
        for image in self.obj.images:
            image.recognize.clean()

    def duplex_copy_image_attr(self, failed_pages, attr, error_msg=None):
        """If in duplex mode, this function will copy the given attribute
        from the image that defines it over to the one that does not.
        ie. if the attribute is None in one and differently in the other image
        it is copied.

        """

        i = 0
        while i < len(self.obj.images):
            failed = (i in failed_pages or i + 1 in failed_pages)

            first = self.obj.images[i]
            second = self.obj.images[i + 1]

            if getattr(first, attr) is None and getattr(second, attr) is None:
                if error_msg is not None and not failed:
                    log.warn(error_msg % (first.filename, first.tiff_page, second.filename, second.tiff_page))
            elif getattr(first, attr) is None:
                setattr(first, attr, getattr(second, attr))
            elif getattr(second, attr) is None:
                setattr(second, attr, getattr(first, attr))

            i += 2


class Image(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'recognize'
    obj_class = model.sheet.Image

    def __init__(self, *args):
        model.buddy.Buddy.__init__(self, *args)

        if self.obj.sheet.survey.defs.style == "classic":
            from . import classic
        elif self.obj.sheet.survey.defs.style == "code128":
            from . import code128
        elif self.obj.sheet.survey.defs.style == "qr":
            from . import qrcode
        elif self.obj.sheet.survey.defs.style == "custom":
            if not hasattr(self.obj, "style"):
                import sys
                log.error(_("No style buddy loaded. This needs to be done for the \"custom\" style!"))
                sys.exit(1)
        else:
            raise AssertionError

    def calculate_rotation(self):
        if self.obj.ignored:
            self.obj.rotated = None
        else:
            self.obj.rotated = self.obj.style.get_page_rotation()

    def calculate_page_number(self):
        if self.obj.ignored:
            self.obj.page_number = None
        else:
            self.obj.page_number = self.obj.style.get_page_number()

    def calculate_survey_id(self):
        if self.obj.ignored:
            self.obj.survey_id = None
        else:
            self.obj.survey_id = self.obj.style.get_survey_id()

    def calculate_questionnaire_id(self):
        if self.obj.ignored:
            self.obj.questionnaire_id = None
        else:
            self.obj.questionnaire_id = self.obj.style.get_questionnaire_id()

    def calculate_global_id(self):
        if self.obj.ignored:
            self.obj.global_id = None
        else:
            self.obj.global_id = self.obj.style.get_global_id()

    def clean(self):
        self.obj.surface.clean()

    def calculate_matrix(self):
        if self.obj.ignored:
            self.obj.matrix.set_px_to_mm(None)
            return

        try:
            # Reset matrix, so that we do not use some existing (and broken)
            # matrix for the resolution estimation.
            self.obj.matrix.set_px_to_mm(None)

            matrix = image.calculate_matrix(
                self.obj.surface.surface,
                self.obj.matrix.mm_to_px(),
                defs.corner_mark_left, defs.corner_mark_top,
                self.obj.sheet.survey.defs.paper_width - defs.corner_mark_left - defs.corner_mark_right,
                self.obj.sheet.survey.defs.paper_height - defs.corner_mark_top - defs.corner_mark_bottom,
            )
        except AssertionError:
            self.obj.matrix.set_px_to_mm(None)
            raise RecognitionError
        else:
            self.obj.matrix.set_px_to_mm(matrix)

    def get_coverage(self, x, y, width, height):
        assert(not self.obj.ignored)

        return image.get_coverage(
            self.obj.surface.surface,
            self.matrix,
            x, y, width, height
        )

    def get_masked_coverage(self, mask, x, y):
        assert(not self.obj.ignored)

        return image.get_masked_coverage(
            self.obj.surface.surface,
            mask,
            x, y
        )

    def get_masked_coverage_without_lines(self, mask, x, y, line_width, line_count):
        assert(not self.obj.ignored)

        return image.get_masked_coverage_without_lines(
            self.obj.surface.surface,
            mask, x, y,
            line_width, line_count
        )

    def get_masked_white_area_count(self, mask, x, y, min_size, max_size):
        assert(not self.obj.ignored)

        return image.get_masked_white_area_count(
            self.obj.surface.surface,
            mask, x, y,
            min_size, max_size
        )

    def correction_matrix_masked(self, x, y, mask):
        assert(not self.obj.ignored)

        return image.calculate_correction_matrix_masked(
            self.obj.surface.surface,
            mask,
            self.matrix,
            x, y
        )

    def find_box_corners(self, x, y, width, height):
        assert(not self.obj.ignored)

        tl, tr, br, bl = image.find_box_corners(
            self.obj.surface.surface,
            self.matrix,
            x, y,
            width, height)

        tolerance = defs.find_box_corners_tolerance
        if(abs(x - tl[0]) > tolerance or
            abs(y - tl[1]) > tolerance or
            abs(x + width - tr[0]) > tolerance or
            abs(y - tr[1]) > tolerance or
            abs(x + width - br[0]) > tolerance or
            abs(y + height - br[1]) > tolerance or
            abs(x - bl[0]) > tolerance or
            abs(y + height - bl[1]) > tolerance
           ):
            raise AssertionError("The found values differ too much from where the box should be.")
        return tl, tr, br, bl

    @property
    def matrix(self):
        return self.obj.matrix.mm_to_px(fallback=False)

class Questionnaire(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'recognize'
    obj_class = model.questionnaire.Questionnaire

    def identify(self, clean=True):
        # recognize image
        try:
            self.obj.sheet.recognize.recognize()
            result = True
        except RecognitionError:
            self.obj.sheet.quality = 0
            result = False

        # clean up
        if clean:
            self.obj.sheet.recognize.clean()

        return result

    def recognize(self):
        # recognize image
        res = self.identify(clean=False)
        if res:
            # iterate over qobjects
            for qobject in self.obj.qobjects:
                qobject.recognize.recognize()

            quality = 1
            for qobject in self.obj.qobjects:
                quality = min(quality, qobject.recognize.get_quality())
            self.obj.sheet.quality = quality

        # Mark the image as "recognized". It might have failed, but even if that
        # happened, we don't want to retry all the time.
        self.obj.sheet.recognized = True
        # Any newly recognized sheet is definately not verified.
        # This is relevant for reruns.
        for img in self.obj.sheet.images:
            img.verified = False

        # clean up
        self.obj.sheet.recognize.clean()


class QObject(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'recognize'
    obj_class = model.questionnaire.QObject

    def recognize(self):
        pass

    def get_quality(self):
        return 1


class Question(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'recognize'
    obj_class = model.questionnaire.Question

    def recognize(self):
        # iterate over boxes
        for box in self.obj.boxes:
            box.recognize.recognize()

    def get_quality(self):
        result = 1
        for box in self.obj.boxes:
            result = min(result, box.data.quality)
        return result


class Box(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'recognize'
    obj_class = model.questionnaire.Box

    def recognize(self):
        pass


class Checkbox(Box, metaclass=model.buddy.Register):

    name = 'recognize'
    obj_class = model.questionnaire.Checkbox

    def prepare_mask(self):
        img = self.obj.sheet.get_page_image(self.obj.page_number)
        width, height = self.obj.width, self.obj.height
        line_width = self.obj.lw

        matrix = list(img.recognize.matrix)
        # Remove any offset from the matrix
        matrix[4] = 0
        matrix[5] = 0
        matrix = cairo.Matrix(*matrix)

        px_width, px_height = matrix.transform_distance(width, height)
        px_width, px_height = int(math.ceil(px_width)), int(math.ceil(px_height))

        surf = cairo.ImageSurface(cairo.FORMAT_A1, px_width, px_height)
        cr = cairo.Context(surf)
        cr.set_source_rgba(0, 0, 0, 0)
        cr.set_operator(cairo.OPERATOR_SOURCE)
        cr.paint()

        # Move to center and apply matrix
        cr.translate(0.5 * px_width, 0.5 * px_height)
        cr.transform(matrix)

        cr.set_source_rgba(0, 0, 0, 1)

        cr.set_line_width(line_width)

        matrix.invert()
        xoff, yoff = matrix.transform_distance(px_width / 2.0, px_height / 2.0)
        xoff = xoff - width / 2
        yoff = yoff - height / 2

        return cr, surf, line_width, width, height, xoff, yoff

    def get_outline_mask(self):
        cr, surf, line_width, width, height, xoff, yoff = self.prepare_mask()

        if self.obj.form == "ellipse":
            cr.save()

            cr.scale((width - line_width) / 2.0, (height - line_width) / 2.0)
            cr.arc(0, 0, 1.0, 0, 2*math.pi)

            # Restore old matrix (without removing the current path)
            cr.restore()
        else:
            cr.translate(-0.5 * width, -0.5 * height)
            cr.rectangle(line_width / 2, line_width / 2, width - line_width / 2, height - line_width / 2)

        cr.stroke()
        surf.flush()
        del cr

        return surf, xoff, yoff

    def get_inner_mask(self):
        """Note this discards half a line width inside the box!"""
        cr, surf, line_width, width, height, xoff, yoff = self.prepare_mask()

        if self.obj.form == "ellipse":
            cr.save()

            cr.scale((width - 3*line_width) / 2.0, (height - 3*line_width) / 2.0)
            cr.arc(0, 0, 1.0, 0, 2*math.pi)

            # Restore old matrix (without removing the current path)
            cr.restore()
        else:
            cr.translate(-0.5 * width, -0.5 * height)
            cr.rectangle(1.5 * line_width, 1.5 * line_width, width - 3 * line_width, height - 3 * line_width)

        cr.fill()
        surf.flush()
        del cr

        return surf, xoff, yoff


    def recognize(self):
        img = self.obj.sheet.get_page_image(self.obj.page_number)

        if img is None or img.recognize.matrix is None:
            self.obj.sheet.valid = 0
            return

        surf, xoff, yoff = self.get_outline_mask()

        matrix, covered = img.recognize.correction_matrix_masked(
            self.obj.x, self.obj.y,
            surf
        )
        # Calculate some sort of quality for the checkbox position
        if covered < defs.image_line_coverage:
            pos_quality = 0
        else:
            pos_quality = min(covered + 0.2, 1)

        x, y = matrix.transform_point(self.obj.x, self.obj.y)
        width, height = matrix.transform_distance(self.obj.width, self.obj.height)
        self.obj.data.x = x + xoff
        self.obj.data.y = y + yoff
        self.obj.data.width = width
        self.obj.data.height = height

        # The debug struct will be filled in if debugging is enabled in the
        # C library. This is done by the boxgallery script currently.
        self.debug = {}

        mask, xoff, yoff = self.get_inner_mask()
        x, y = self.obj.data.x, self.obj.data.y
        x, y = x + xoff, y + yoff

        x, y = img.recognize.matrix.transform_point(x, y)
        x, y = int(x), int(y)

        # This is not the outline, but the width of the drawn stroke!
        remove_line_width = 1.2 * pt_to_mm
        remove_line_width_px = max(img.recognize.matrix.transform_distance(remove_line_width, remove_line_width))

        coverage = img.recognize.get_masked_coverage(mask, x, y)
        self.obj.data.metrics['coverage'] = coverage
        self.debug['coverage'] = image.get_debug_surface()

        # Remove 3 lines with width 1.2pt(about 5px).
        coverage = img.recognize.get_masked_coverage_without_lines(mask, x, y, remove_line_width_px, 3)
        self.obj.data.metrics['cov-lines-removed'] = coverage
        self.debug['cov-lines-removed'] = image.get_debug_surface()

        count, coverage = img.recognize.get_masked_white_area_count(mask, x, y, 0.05, 1.0)
        self.obj.data.metrics['cov-min-size'] = coverage
        self.debug['cov-min-size'] = image.get_debug_surface()

        state = 0
        quality = -1
        # Iterate the ranges
        for metric, value in self.obj.data.metrics.items():
            metric = defs.checkbox_metrics[self.obj.sheet.survey.defs.checkmode][metric]

            for lower, upper in zip(metric[:-1], metric[1:]):
                if value >= lower[0] and value <= upper[0]:
                    # Interpolate quality value
                    if lower[0] != upper[0]:
                        metric_quality = lower[2] + (upper[2] - lower[2]) * (value - lower[0]) / (upper[0] - lower[0])
                    else:
                        metric_quality = lower[2]

                    if metric_quality > quality:
                        state = lower[1]
                        quality = metric_quality

        self.obj.data.state = state
        self.obj.data.quality = min(quality, pos_quality)


class Textbox(Box, metaclass=model.buddy.Register):

    name = 'recognize'
    obj_class = model.questionnaire.Textbox

    def recognize(self):
        class Quadrilateral():
            """This class iterates through small box areas in a quadriliteral.
            This is usefull because some scanners have trapezoidal distortions."""
            # Assumes top left, top right, bottom right, bottom left
            # corner.
            def __init__(self, p0, p1, p2, p3):
                self.x0 = p0[0]
                self.y0 = p0[1]
                self.x1 = p1[0]
                self.y1 = p1[1]
                self.x2 = p2[0]
                self.y2 = p2[1]
                self.x3 = p3[0]
                self.y3 = p3[1]

                # 0 -> 1
                self.m0 = (self.y1 - self.y0) / (self.x1 - self.x0)
                self.m1 = (self.x2 - self.x1) / (self.y2 - self.y1)
                self.m2 = (self.y3 - self.y2) / (self.x3 - self.x2)
                self.m3 = (self.x0 - self.x3) / (self.y0 - self.y3)

                self.top = min(self.y0, self.y1)
                self.bottom = max(self.y2, self.y3)
                self.left = min(self.x0, self.x3)
                self.right = max(self.x1, self.x2)

            def iterate_bb(self, step_x, step_y, test_width, test_height, padding):
                y = self.top
                while y + test_height < self.bottom:
                    x = self.left
                    while x + test_width < self.right:
                        yield x, y
                        x += step_x

                    y += step_y

            def iterate_outline(self, step_x, step_y, test_width, test_height, padding):
                # Top
                x, y = self.x0, self.y0
                x += padding
                y += padding

                dest_x = self.x1 - padding - test_width
                dest_y = self.y1 + padding

                dist_x = dest_x - x
                dist_y = dest_y - y

                length = math.sqrt(dist_x ** 2 + dist_y ** 2)
                for step in range(int(length / step_x)):
                    yield x + dist_x * step / (length / step_x), y + dist_y * step / (length / step_x)
                yield dest_x, dest_y

                # Bottom
                x, y = self.x3, self.y3
                x = x + padding
                y = y - padding - test_height

                dest_x = self.x2 - padding - test_width
                dest_y = self.y2 - padding - test_height

                dist_x = dest_x - x
                dist_y = dest_y - y

                length = math.sqrt(dist_x ** 2 + dist_y ** 2)
                for step in range(int(length / step_x)):
                    yield x + dist_x * step / (length / step_x), y + dist_y * step / (length / step_x)
                yield dest_x, dest_y

                # Left
                x, y = self.x0, self.y0
                x += padding
                y += padding

                dest_x = self.x3 + padding
                dest_y = self.y3 - padding - test_height

                dist_x = dest_x - x
                dist_y = dest_y - y

                length = math.sqrt(dist_x ** 2 + dist_y ** 2)
                for step in range(int(length / step_y)):
                    yield x + dist_x * step / (length / step_y), y + dist_y * step / (length / step_y)
                yield dest_x, dest_y

                # Right
                x, y = self.x1, self.y1
                x = x - padding - test_width
                y = y + padding

                dest_x = self.x2 - padding - test_width
                dest_y = self.y2 - padding - test_height

                dist_x = dest_x - x
                dist_y = dest_y - y

                length = math.sqrt(dist_x ** 2 + dist_y ** 2)
                for step in range(int(length / step_y)):
                    yield x + dist_x * step / (length / step_y), y + dist_y * step / (length / step_y)
                yield dest_x, dest_y

            def iterate(self, step_x, step_y, test_width, test_height, padding):
                for x, y in self.iterate_bb(step_x, step_y, test_width, test_height, padding):
                    ly = self.y0 + self.m0 * (x - self.x0)
                    if not ly + padding < y:
                        continue

                    ly = self.y2 + self.m2 * (x - self.x2)
                    if not ly - padding > y + test_height:
                        continue

                    lx = self.x1 + self.m1 * (y - self.y1)
                    if not lx - padding > x + test_width:
                        continue

                    lx = self.x3 + self.m3 * (y - self.y3)
                    if not lx + padding < x:
                        continue

                    yield x, y

                for x, y in self.iterate_outline(step_x, step_y, test_width, test_height, padding):
                    yield x, y

        bbox = None
        img = self.obj.sheet.get_page_image(self.obj.page_number)

        if img is None or img.recognize.matrix is None:
            self.obj.sheet.valid = 0
            return

        x = self.obj.x
        y = self.obj.y
        width = self.obj.width
        height = self.obj.height

        # Scanning area and stepping
        step_x = defs.textbox_scan_step_x
        step_y = defs.textbox_scan_step_x
        test_width = defs.textbox_scan_width
        test_height = defs.textbox_scan_height

        # extra_padding is always added to the box side at the end.
        extra_padding = defs.textbox_extra_padding
        scan_padding = defs.textbox_scan_uncorrected_padding

        quad = Quadrilateral((x, y), (x + width, y), (x + width, y + height), (x, y + height))
        try:
            quad = Quadrilateral(*img.recognize.find_box_corners(x, y, width, height))
            # Lower padding, as we found the corners and are therefore more acurate
            scan_padding = defs.textbox_scan_padding
        except AssertionError:
            pass

        surface = img.surface.surface
        matrix = img.recognize.matrix
        for x, y in quad.iterate(step_x, step_y, test_width, test_height, scan_padding):
            # Use the image module directly as we are calling in *a lot*
            coverage = image.get_coverage(surface, matrix, x, y, test_width, test_height)
            if coverage > defs.textbox_scan_coverage:
                if not bbox:
                    bbox = [x, y, test_width, test_height]
                else:
                    bbox_x = min(bbox[0], x)
                    bbox_y = min(bbox[1], y)
                    bbox[2] = max(bbox[0] + bbox[2], x + test_width) - bbox_x
                    bbox[3] = max(bbox[1] + bbox[3], y + test_height) - bbox_y
                    bbox[0] = bbox_x
                    bbox[1] = bbox_y

        if bbox and (bbox[2] > defs.textbox_minimum_writing_width or
                     bbox[3] > defs.textbox_minimum_writing_height):
            # Do not accept very small bounding boxes.
            self.obj.data.state = True

            self.obj.data.x = bbox[0] - (scan_padding + extra_padding)
            self.obj.data.y = bbox[1] - (scan_padding + extra_padding)
            self.obj.data.width = bbox[2] + 2 * (scan_padding + extra_padding)
            self.obj.data.height = bbox[3] + 2 * (scan_padding + extra_padding)
        else:
            self.obj.data.state = False

            self.obj.data.x = self.obj.x
            self.obj.data.y = self.obj.y
            self.obj.data.width = self.obj.width
            self.obj.data.height = self.obj.height

class Codebox(Textbox, metaclass=model.buddy.Register):

    name = 'recognize'
    obj_class = model.questionnaire.Codebox

    def recognize(self):
        img = self.obj.sheet.get_page_image(self.obj.page_number)

        if img is None or img.recognize.matrix is None:
            self.obj.sheet.valid = 0
            return

        x = self.obj.x
        y = self.obj.y
        width = self.obj.width
        height = self.obj.height

        surface = img.surface.surface
        matrix = img.recognize.matrix

        res = read_barcode(surface, matrix,
                           x, y, width, height,
                           "QRCODE")

        # This is maybe a bit odd, but we accept empty string as a valid
        # content (if there was a barcode which was just the empty string).
        if res is not None:
            self.obj.data.state = True
            self.obj.data.text = res
        else:
            self.obj.data.state = False
            self.obj.data.text = None
